with Util.Tasking.Complex_Locks;
with Ada.Text_IO;
with Ada.Task_Identification;

procedure Test_RWLock
is
   use Ada.Text_IO;
   use Util.Tasking.Complex_Locks;

   use type Ada.Task_Identification.Task_Id;

   Had_A_Failure : Boolean := False;

   procedure Failure
     (Msg : in String)
   is
   begin
      Had_A_Failure := True;
      Put_Line ("Failure: " & Msg);
   end Failure;

   procedure Success
     (Msg : in String)
   is
   begin
      Put_Line ("Success: " & Msg);
   end Success;

   type Status is (Idle, Read_Request, Write_Request, Lock_Granted,
                   Releasing, Released, Demoting, Demoted, Failed);

   protected type Task_Status is
      procedure Requesting_Read;
      procedure Requesting_Write;
      procedure Lock_Gotten (Mode : in Lock_Holder_Mode);
      procedure Status_Now (S : out Status;
                            Mode : out Lock_Holder_Mode;
                            N : out Natural);
      entry Wait (S : out Status;
                  Mode : out Lock_Holder_Mode;
                  N : out Natural);
      procedure Failure;
      procedure Release_Request;
      procedure Release_Done (Mode : in Lock_Holder_Mode);
      procedure Demote_Request;
      procedure Demote_Done (Mode : in Lock_Holder_Mode);
   private
      Stat    : Status           := Idle;
      Holding : Lock_Holder_Mode := Not_Holder;
      Count   : Natural          := 0;
      Changed : Boolean          := False;
   end Task_Status;

   protected body Task_Status is

      procedure Requesting_Read
      is
      begin
         Stat := Read_Request;
         Changed := True;
      end Requesting_Read;

      procedure Requesting_Write
      is
      begin
         Stat := Write_Request;
         Changed := True;
      end Requesting_Write;

      procedure Demote_Request
      is
      begin
         Stat    := Demoting;
         Changed := True;
      end Demote_Request;

      procedure Demote_Done (Mode : in Lock_Holder_Mode)
      is
      begin
         Stat    := Demoted;
         Holding := Mode;
         Changed := True;
      end Demote_Done;

      procedure Release_Request
      is
      begin
         Stat    := Releasing;
         Changed := True;
      end Release_Request;

      procedure Release_Done (Mode : in Lock_Holder_Mode)
      is
      begin
         Stat    := Released;
         Holding := Mode;
         Count   := Count - 1;
         Changed := True;
      end Release_Done;

      procedure Failure
      is
      begin
         Stat := Failed;
         Changed := True;
      end Failure;

      procedure Lock_Gotten (Mode : in Lock_Holder_Mode)
      is
      begin
         Stat    := Lock_Granted;
         Holding := Mode;
         Count   := Count + 1;
         Changed := True;
      end Lock_Gotten;

      procedure Status_Now
        (S    :    out Status;
         Mode :    out Lock_Holder_Mode;
         N    :    out Natural)
      is
      begin
         S := Stat; Mode := Holding; N := Count;
      end Status_Now;

      entry Wait
        (S    :    out Status;
         Mode :    out Lock_Holder_Mode;
         N    :    out Natural)
        when Changed
      is
      begin
         S := Stat; Mode := Holding; N := Count;
         if Wait'Count = 0 then Changed := False; end if;
      end Wait;

   end Task_Status;

   task type Tester_Task (Stat : access Task_Status;
                          Lock : access Task_Read_Write_Lock)
   is
      entry Stop;
      entry Request (Mode : in Lock_Mode);
      entry Release;
      entry Demote;
   end Tester_Task;

   task body Tester_Task
   is
      The_Mode : Lock_Mode;
   begin
      loop
         select
            accept Stop;
            exit;
         or
            accept Request (Mode : in Lock_Mode) do
               The_Mode := Mode;
               if The_Mode = Shared then
                  Stat.Requesting_Read;
               else
                  Stat.Requesting_Write;
               end if;
            end Request;
            declare
               Real_Mode : Lock_Holder_Mode;
            begin
               Put_Line ("Requesting " & Lock_Mode'Image (The_Mode));
               Lock.Acquire (The_Mode);
               Put_Line ("Holding...");
               Lock.Holding (Real_Mode);
               Put_Line ("Holding " & Lock_Holder_Mode'Image (Real_Mode));
               Stat.Lock_Gotten (Real_Mode);
            exception
               when others =>
                  Put_Line ("Failed");
                  Stat.Failure;
            end;
         or
            accept Release do
               Stat.Release_Request;
            end Release;
            begin
               Lock.Release;
               declare
                  Real_Mode : Lock_Holder_Mode;
               begin
                  Lock.Holding (Real_Mode);
                  Stat.Release_Done (Real_Mode);
               end;
            exception
               when others =>
                  Stat.Failure;
            end;
         or
            accept Demote do
              Stat.Demote_Request;
            end Demote;
            Lock.Demote;
            declare
               Real_Mode : Lock_Holder_Mode;
            begin
               Lock.Holding (Real_Mode);
               Stat.Demote_Done (Real_Mode);
            end;
         or
            terminate;
         end select;
      end loop;
   end Tester_Task;

   type T_Ptr is access all Tester_Task;

   type Tester is limited
      record
         Stat : aliased Task_Status;
         T    : T_Ptr;
      end record;

   Simple_Test : Tester;
   Many_Tasks  : array (Positive range 1 .. 10) of Tester;
   The_Lock    : aliased Task_Read_Write_Lock;

   Current_Status : Status;
   Current_Mode : Lock_Holder_Mode;
   Current_N : Natural;

   procedure Expect
     (Stat : in Status;
      Mode : in Lock_Holder_Mode;
      N    : in Natural;
      Msg  : in String := "")
   is
   begin
      if Current_Status /= Stat or else
         Current_Mode /= Mode or else
         Current_N /= N
      then
         Failure (Msg & ' ' &
                  Status'Image (Current_Status) & ' ' &
                  Lock_Holder_Mode'Image (Current_Mode) & ' ' &
                  Natural'Image (Current_N) &
                  " (expected " &
                  Status'Image (Stat) & ' ' &
                  Lock_Holder_Mode'Image (Mode) & ' ' &
                  Natural'Image (N) &
                  ')');
      else
         Success (Msg & ' ' &
                  Status'Image (Current_Status) & ' ' &
                  Lock_Holder_Mode'Image (Current_Mode) & ' ' &
                  Natural'Image (Current_N));
      end if;
   end Expect;

begin
   begin
      Simple_Test.T :=
        new Tester_Task (Simple_Test.Stat'Access, The_Lock'Access);
      Simple_Test.T.Request (Shared);
      Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      if Current_Status = Read_Request then
         Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Lock_Granted, Shared, 1);
      Simple_Test.T.Request (Shared);
      Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      if Current_Status = Read_Request then
         Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Lock_Granted, Shared, 2);
      Simple_Test.T.Release;
      Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      if Current_Status = Releasing then
         Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Released, Shared, 1);
      Simple_Test.T.Request (Exclusive);
      Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      if Current_Status = Write_Request then
         Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Lock_Granted, Exclusive, 2);
      Simple_Test.T.Request (Shared);
      Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      if Current_Status = Read_Request then
         Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Lock_Granted, Exclusive, 3);
      Simple_Test.T.Release;
      Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      if Current_Status = Releasing then
         Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Released, Exclusive, 2);
      Simple_Test.T.Demote;
      Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      if Current_Status = Demoting then
         Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Demoted, Shared, 2);
      Simple_Test.T.Release;
      Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      if Current_Status = Releasing then
         Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Released, Shared, 1);
      Simple_Test.T.Release;
      Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      if Current_Status = Releasing then
         Simple_Test.Stat.Wait (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Released, Not_Holder, 0);
      Simple_Test.T.Stop;
   end;
   if Had_A_Failure then Failure ("stopping tests."); return; end if;
   --  Ok, it works at least for one task. Now try with several tasks.
   if Many_Tasks'Length = 1 then return; end if;
   declare
      Last_Reader : Natural;
   begin
      Last_Reader := (Many_Tasks'Length * 7) / 10 + Many_Tasks'First - 1;
      if Many_Tasks'Length < 10 or else
         Last_Reader + 1 >= Many_Tasks'Last or else
         Last_Reader <= Many_Tasks'First
      then
         Failure ("not enough tasks in Many_Tasks");
         return;
      end if;
      for I in Many_Tasks'Range loop
         Many_Tasks (I).T := new Tester_Task (Many_Tasks (I).Stat'Access,
                                              The_Lock'Access);
      end loop;
      --  We have at least 7 readers and 3 writers.
      for I in Many_Tasks'First .. Last_Reader - 2 loop
         Many_Tasks (I).T.Request (Shared);
         Many_Tasks (I).Stat.Wait (Current_Status, Current_Mode, Current_N);
         if Current_Status = Read_Request then
            Many_Tasks (I).Stat.Wait (Current_Status, Current_Mode, Current_N);
         end if;
         Expect (Lock_Granted, Shared, 1, "Task #" & Natural'Image (I));
      end loop;
      --  Now a writer.
      Many_Tasks (Last_Reader + 1).T.Request (Exclusive);
      delay 2.0;
      Many_Tasks (Last_Reader + 1).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      Expect (Write_Request, Not_Holder, 0,
              "Task #" & Natural'Image (Last_Reader + 1));
      --  And the last two readers.
      for I in Last_Reader - 1 ..Last_Reader loop
         Many_Tasks (I).T.Request (Shared);
         delay 2.0;
         Many_Tasks (I).Stat.Wait (Current_Status, Current_Mode, Current_N);
         Expect (Read_Request, Not_Holder, 0, "Task #" & Natural'Image (I));
      end loop;
      declare
         N : Natural := The_Lock.Queued;
      begin
         if N /= 3 then
            Failure (Natural'Image (N) & " tasks queued, expected 3");
         end if;
      end;
      for I in Many_Tasks'First .. Last_Reader - 4 loop
         Many_Tasks (I).T.Request (Shared);
         Many_Tasks (I).Stat.Wait (Current_Status, Current_Mode, Current_N);
         if Current_Status = Read_Request then
            Many_Tasks (I).Stat.Wait (Current_Status, Current_Mode, Current_N);
         end if;
         Expect (Lock_Granted, Shared, 2, "Task #" & Natural'Image (I));
      end loop;
      --  Now a promotion (which will be queued)
      Many_Tasks (Last_Reader - 2).T.Request (Exclusive);
      delay 2.0;
      Many_Tasks (Last_Reader - 2).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      Expect (Write_Request, Shared, 1,
              "Task #" & Natural'Image (Last_Reader - 2));
      --  Now a second promotion (which should fail)
      Many_Tasks (Last_Reader - 3).T.Request (Exclusive);
      delay 2.0;
      Many_Tasks (Last_Reader - 3).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      if Current_Status = Write_Request then
         Many_Tasks (Last_Reader - 3).Stat.Wait
           (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Failed, Shared, 1,
              "Task #" & Natural'Image (Last_Reader - 3));
      --  Now release all locks
      for I in Many_Tasks'First .. Last_Reader - 3 loop
         Many_Tasks (I).T.Release;
         Many_Tasks (I).Stat.Wait (Current_Status, Current_Mode, Current_N);
         if Current_Status = Releasing then
            Many_Tasks (I).Stat.Wait (Current_Status, Current_Mode, Current_N);
         end if;
         if I /= Last_Reader - 3 then
            Expect (Released, Shared, 1, "Task #" & Natural'Image (I));
            Many_Tasks (I).T.Release;
            Many_Tasks (I).Stat.Wait (Current_Status, Current_Mode, Current_N);
            if Current_Status = Releasing then
               Many_Tasks (I).Stat.Wait
                 (Current_Status, Current_Mode, Current_N);
            end if;
         end if;
         Expect (Released, Not_Holder, 0, "Task #" & Natural'Image (I));
      end loop;
      delay 2.0;
      --  Now the waiting promotion should have gotten the lock:
      Many_Tasks (Last_Reader - 2).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      Expect (Lock_Granted, Exclusive, 2,
              "Task #" & Natural'Image (Last_Reader - 2));
      if The_Lock.Writer /= Many_Tasks (Last_Reader - 2).T'Identity then
         Failure ("wrong TID for writer, task #" &
                  Natural'Image (Last_Reader - 2));
      end if;
      Many_Tasks (Last_Reader - 2).T.Release;
      Many_Tasks (Last_Reader - 2).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      if Current_Status = Releasing then
         Many_Tasks (Last_Reader - 2).Stat.Wait
           (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Released, Exclusive, 1,
              "Task #" & Natural'Image (Last_Reader - 2));
      Many_Tasks (Last_Reader - 2).T.Release;
      Many_Tasks (Last_Reader - 2).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      if Current_Status = Releasing then
         Many_Tasks (Last_Reader - 2).Stat.Wait
           (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Released, Not_Holder, 0,
              "Task #" & Natural'Image (Last_Reader - 2));
      --  Now the writer should have gotten the lock.
      Many_Tasks (Last_Reader + 1).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      Expect (Lock_Granted, Exclusive, 1,
              "Task #" & Natural'Image (Last_Reader + 1));
      --  And we still should have two queued tasks
      declare
         N : Natural := The_Lock.Queued;
      begin
         if N /= 2 then
            Failure (Natural'Image (N) & " tasks queued, expected 2");
         end if;
      end;
      --  Queue another writer.
      Many_Tasks (Last_Reader + 2).T.Request (Exclusive);
      delay 2.0;
      Many_Tasks (Last_Reader + 2).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      Expect (Write_Request, Not_Holder, 0,
              "Task #" & Natural'Image (Last_Reader + 2));
      --  Release the exclusive lock.
      Many_Tasks (Last_Reader + 1).T.Release;
      Many_Tasks (Last_Reader + 1).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      if Current_Status = Releasing then
         Many_Tasks (Last_Reader + 1).Stat.Wait
           (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Released, Not_Holder, 0,
              "Task #" & Natural'Image (Last_Reader + 1));
      --  The two readers should have gotten the lock.
      Many_Tasks (Last_Reader - 1).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      Expect (Lock_Granted, Shared, 1,
              "Task #" & Natural'Image (Last_Reader - 1));
      Many_Tasks (Last_Reader).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      Expect (Lock_Granted, Shared, 1,
              "Task #" & Natural'Image (Last_Reader));
      --  The writer should still be waiting.
      Many_Tasks (Last_Reader + 2).Stat.Status_Now
        (Current_Status, Current_Mode, Current_N);
      Expect (Write_Request, Not_Holder, 0,
              "Task #" & Natural'Image (Last_Reader + 2));
      -- And there should be only one task queued now.
      declare
         N : Natural := The_Lock.Queued;
      begin
         if N /= 1 then
            Failure (Natural'Image (N) & " tasks queued, expected 2");
         end if;
      end;
      Many_Tasks (Last_Reader - 1).T.Release;
      Many_Tasks (Last_Reader).T.Release;
      --  The two readers should have released the lock.
      Many_Tasks (Last_Reader - 1).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      Expect (Released, Not_Holder, 0,
              "Task #" & Natural'Image (Last_Reader - 1));
      Many_Tasks (Last_Reader).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      Expect (Released, Not_Holder, 0,
              "Task #" & Natural'Image (Last_Reader));
      --  And the writer should have gotten it.
      Many_Tasks (Last_Reader + 2).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      Expect (Lock_Granted, Exclusive, 1,
              "Task #" & Natural'Image (Last_Reader + 2));
      Many_Tasks (Last_Reader + 2).T.Release;
      Many_Tasks (Last_Reader + 2).Stat.Wait
        (Current_Status, Current_Mode, Current_N);
      if Current_Status = Releasing then
         Many_Tasks (Last_Reader + 2).Stat.Wait
           (Current_Status, Current_Mode, Current_N);
      end if;
      Expect (Released,  Not_Holder, 0,
              "Task #" & Natural'Image (Last_Reader + 2));
      -- And there should be no task queued now.
      declare
         N : Natural := The_Lock.Queued;
      begin
         if N /= 0 then
            Failure (Natural'Image (N) & " tasks queued, expected 2");
         end if;
      end;
   end;
   if Had_A_Failure then
      Failure ("Some tests failed.");
   else
      Success ("All tests succeeded.");
   end if;
end Test_RWLock;
