-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2001, 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    UTC and daylight saving time utility functions.
--  </PURPOSE>
--
--  <TASKING>
--    Fully task safe.
--  </TASKING>
--
--  <NO_STORAGE>
--
--  <HISTORY>
--    28-OCT-2002   TW  Initial version.
--  </HISTORY>
-------------------------------------------------------------------------------

with Ada.Calendar;
with Interfaces.C;

with Util.Fixed;

package body Util.Calendar.UTC is

   package Fixed_Ops is new Util.Fixed (Ada.Calendar.Day_Duration);

   package C_Interface is

      type C_Time_Type is
         record
            Year    : Interfaces.C.int; --  Gregorian, >= 1900
            Month   : Interfaces.C.int; --  1 .. 12
            Day     : Interfaces.C.int; --  1 .. 31
            Hours   : Interfaces.C.int; --  0 .. 23
            Minutes : Interfaces.C.int; --  0 .. 59
            Seconds : Interfaces.C.int; --  0 .. 61 (!)
         end record;
      pragma Convention (C, C_Time_Type);

      function UTC (Now : access C_Time_Type) return Interfaces.C.int;
      pragma Import (C, UTC, "util_ada_get_utc");

      function Is_DST return Interfaces.C.int;
      pragma Import (C, Is_DST, "util_ada_is_dst");

      --  These two functions come from file 'util-wrappers.c', which is
      --  included in 'libutil.a'.

   end C_Interface;

   --  The underlying C functions 'gmtime' and 'localtime' are not reentrant:
   --  they return a pointer to a global variable! (Sorry, I'm not responsible
   --  for the C "gurus" having botched their standard so badly.) Anyway, this
   --  means we have to ensure that no two tasks ever call the C wrappers
   --  simultaneously, and hence we wrap the accesses in a protected object.

   protected C_Time is
      procedure Get_UTC (Val : out C_Interface.C_Time_Type);
      procedure Get_DST (Val : out Boolean);
   end C_Time;

   protected body C_Time is

      procedure Get_UTC (Val : out C_Interface.C_Time_Type)
      is
         use type Interfaces.C.int;
         Tmp    : aliased C_Interface.C_Time_Type;
         Status : Interfaces.C.int := C_Interface.UTC (Tmp'Access);
      begin
         if Status < 0 then raise Ada.Calendar.Time_Error; end if;
         Val := Tmp;
      end Get_UTC;

      procedure Get_DST (Val : out Boolean)
      is
         use type Interfaces.C.int;
         Status : Interfaces.C.int := C_Interface.Is_DST;
      begin
         if Status < 0 then raise Ada.Calendar.Time_Error; end if;
         Val := Status > 0;
      end Get_DST;

   end C_Time;

   function UTC return Ada.Calendar.Time
   is
      use Ada.Calendar;
      use type Interfaces.C.int;

      Tmp : C_Interface.C_Time_Type;

   begin
      C_Time.Get_UTC (Tmp);
      declare
         Secs : Day_Duration :=
           Day_Duration (Tmp.Hours * 3600 + Tmp.Minutes * 60);
      begin
         if Tmp.Seconds >= 60 then
            Secs := Secs + 60.0;
         else
            Secs := Secs + Day_Duration (Tmp.Seconds);
         end if;
         if Secs < Day_Duration'Last then
            --  Fill in the fraction. (UTC is always offset by a
            --  non-fractional number of seconds.)
            Secs := Secs + Fixed_Ops.Fraction (Seconds (Clock));
         end if;
         return Time_Of (Year_Number (Tmp.Year),
                         Month_Number (Tmp.Month),
                         Day_Number (Tmp.Day),
                         Secs);
      exception
         when others =>
            raise Ada.Calendar.Time_Error;
      end;
   end UTC;

   function Is_DST return Boolean
   is
      Tmp : Boolean;
   begin
      C_Time.Get_DST (Tmp);
      return Tmp;
   end Is_DST;

end Util.Calendar.UTC;
