-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2001, 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    Provides a Gregorian calendar. The Gregorian calendar is the one that
--    is used throughout the world today. It had been introduced by the papal
--    bull of pope Gregory XIII on 24-FEB-1582 (Julian!). This bull decreed
--    that the day following 4-OCT-1582 (Julian) would be 15-OCT-1582.
--    The latter is thus the base date for the Gregorian calendar (Julian day
--    number 2299161).
--
--    As an aside, please note that the weekday was left unaffected by this
--    change! 4-OCT-1582 was a Thursday, and the next day was a Friday,
--    irrespective of the calendar used!
--
--    The Gregorian calendar differs from the Julian calendar in the way leap
--    years are determined and in the way the easter date is calculated.
--
--    Not all countries and regions changed that day from the Julian to the
--    new Gregorian calendars. Many countries adopted the Gregorian calendar
--    only in the 20th century! This implementation ignores all these
--    differences; if you need a localized calendar that takes account of
--    all the very complicated and illogical calendarical changes as they
--    really happened, you can always construct your own calendar from
--    the Julian and Gregorian ones. See the Calendar FAQ for some change
--    dates. Also see <CODE>
--    <A HREF="util-calendars-universal.html">Util.Calendars.Universal</A>
--    </CODE> for a calendar that combines the Julian and the Gregorian
--    calendars, and that allows you to define the date the swicth occurred.
--
--    This is <EM>not</EM> a proleptic calendar: trying to use it on dates
--    before Julian day 2299161 will result in a @Date_Error@ exception.
--  </PURPOSE>
--
--  <NOT_TASK_SAFE>
--
--  <NO_STORAGE>
--
--  <HISTORY>
--    03-OCT-2001   TW  Initial version.
--  </HISTORY>
-------------------------------------------------------------------------------

with Util.Calendars.Julian;

package body Util.Calendars.Gregorian is

   use Util.Times;
   use Util.Calendars.Western;

   Base_Year : constant Year_Number := 1582;

   Base_Day  : Julian_Day_Number := 2299161; --  15-OCT-1582 Gregorian.

   function Is_Leap_Year
     (Base : in Calendar;
      Year : in Year_Number)
     return Boolean
   is
      pragma Warnings (Off, Base); --  silence -gnatwa
   begin
      return (Year mod 4 = 0 and then Year mod 100 /= 0) or else
             Year mod 400 = 0;
   end Is_Leap_Year;

   ----------------------------------------------------------------------------
   --  Conversion routines.

   function Julian_Day
     (Base  : in Calendar;
      Year  : in Year_Number;
      Month : in Month_Number;
      Day   : in Day_Number)
     return Julian_Day_Number
   is
      pragma Warnings (Off, Base); --  silence -gnatwa
   begin
      if Year < 1582 or else
         (Year = 1582 and then
          (Month < 10 or else (Month = 10 and then Day < 15)))
      then
         raise Date_Error;
      end if;
      declare
         A : Integer := (14 - Integer (Month)) / 12;
         Y : Integer := Integer (Year) + 4800 - A;
         M : Integer := Integer (Month) + 12 * A - 3;
      begin
         return
           Julian_Day_Number (Integer (Day) +
                              (153 * M + 2) / 5 +
                              Y * 365 + Y / 4 - Y / 100 + Y / 400 -
                              32045);
      end;
   end Julian_Day;

   procedure Date
     (Base       : in     Calendar;
      Julian_Day : in     Julian_Day_Number;
      Year       :    out Year_Number;
      Month      :    out Month_Number;
      Day        :    out Day_Number)
   is
      pragma Warnings (Off, Base); --  silence -gnatwa
   begin
      if Julian_Day < Base_Day then raise Date_Error; end if;
      declare
         A : Integer := Integer (Julian_Day) + 32044;
         B : Integer := (4 * A + 3) / 146097;
         C : Integer := A - (B * 146097) / 4;
         D : Integer := (4 * C + 3) / 1461;
         E : Integer := C - (1461 * D) / 4;
         M : Integer := (5 * E + 2) / 153;
      begin
         Day   := Day_Number (E - (153 * M + 2) / 5 + 1);
         Month := Month_Number (M + 3 - 12 * (M / 10));
         Year  := Year_Number (B * 100 + D - 4800 + M / 10);
      end;
   end Date;

   ----------------------------------------------------------------------------
   --  A system in which days are counted.

   function Day_In_Year
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Day_Count
   is
   begin
      if Julian_Day (Date) < Base_Day then raise Date_Error; end if;
      if Julian_Day (Date) < Julian_Day (Base, 1583, 1, 1) then
         --  Calculate using the Julian calendar and subtract 10.
         declare
            J : Util.Calendars.Julian.Calendar;
         begin
            return Util.Calendars.Julian.Day_In_Year (J, Date) - 10;
         end;
      else
         --  Super call.
         return Day (Util.Calendars.Western.Calendar (Base), Date);
      end if;
   end Day_In_Year;

   function Time_Of
     (Base    : in Calendar;
      Year    : in Year_Number;
      Day     : in Day_Count;
      Seconds : in Day_Duration := 0.0)
     return Util.Times.Time
   is
   begin
      if Day = 366 and then not Is_Leap_Year (Base, Year) then
         raise Date_Error;
      end if;
      if Year = 1582 then
         --  That year had only 355 days, and day #278 was Oct 15.
         if Day < 278 or else Day > 355 then
            raise Date_Error;
         end if;
         --  Use the Julian calendar, but don't add 10!
         declare
            J : Util.Calendars.Julian.Calendar;
         begin
            return Util.Calendars.Julian.Time_Of (J, Year, Day);
         end;
      else
         return
           Time_Of (Julian_Day (Base, Year, 1, 1) +
                    Julian_Day_Number (Day) - 1,
                    Seconds);
      end if;
   end Time_Of;

   ----------------------------------------------------------------------------
   --  Easter date

   function Easter
     (Base : in Calendar;
      Year : in Year_Number)
     return Util.Times.Time
   is
   begin
      if Year <= Base_Year then raise Date_Error; end if;
      declare
         Y : Integer := Integer (Year);
         G : Integer := Y mod 19;
         C : Integer := Y / 100;
         H : Integer := (C - C / 4 - (8 * C + 13) / 25 + 19 * G + 15) mod 30;
         I : Integer :=
           H - (H / 28) * (1 - (H / 28) * (29 / (H + 1)) * ((21 - G) / 11));
         J : Integer := (Y + Y / 4 + I + 2 - C + C / 4) mod 7;
         L : Integer := I - J;
         M : Integer := 3 + (L + 40) / 44;
      begin
         return Time_Of (Base, Year, Month_Number (M),
                         Day_Number (L + 28 - 31 * (M / 4)));
      end;
   end Easter;

   ----------------------------------------------------------------------------
   --  ISO 8601 week calculations. Weeks always start on a Monday.

   function Day_Of_Week
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Week_Day
   is
      Year  : Year_Number;
      Month : Month_Number;
      Day   : Day_Number;
   begin
      Split (Base, Date, Year, Month, Day);
      declare
         A : Integer := (14 - Integer (Month)) / 12;
         Y : Integer := Integer (Year) - A;
         M : Integer := Integer (Month) + 12 * A - 2;
         D : Integer := (Integer (Day) +
                         Y + Y / 4 - Y / 100 + Y / 400 +
                         (31 * M) / 12) mod 7;
      begin
         --  Now D=0 means Sunday...
         D := (D + 6) mod 7;
         --  Now we've mapped it such that D=0 is Monday and D=6 is Sunday.
         return Week_Day'Val (D);
      end;
   end Day_Of_Week;

   procedure Week
     (Base : in     Calendar;
      Date : in     Util.Times.Time;
      Year :    out Year_Number;
      Week :    out Week_Number)
   is
      Month : Month_Number;
      Day   : Day_Number;
   begin
      Split (Base, Date, Year, Month, Day);
      declare
         J : Integer := Julian_Day (Date);
         D : Integer :=
           (((J + 31741 - J mod 7) mod 146097) mod 36524) mod 1461;
         L : Integer := D / 1460;
         E : Integer := ((D - L) mod 365) + L;
      begin
         Week := Week_Number (E / 7 + 1);
         --  Now adjust the year for dates in late December/early January,
         --  where we may be in the first or last week of the following or
         --  preceding year.
         if Month = 1 and then Week >= 52 then
            Year := Year - 1;
         elsif Month = 12 and then Week <= 1 then
            Year := Year + 1;
         end if;
      end;
   end Week;

end Util.Calendars.Gregorian;
