-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2001, 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    This is a proleptic Julian calendar. The Julian calendar had been
--    introduced by Julius Caesar in 46 B.C. and been in use since 45 B.C.
--
--    This is a proleptic calendar: it extends to dates before 45 B.C.
--
--    Although every fourth year in the Julian calendar is a leap year, a
--    counting error made every third year between 45 B.C. and 9 B.C. a
--    leap year, which was corrected by not having leap years from 8 B.C.
--    until 8 A.D., which again was a leap year. This implementation does
--    <EM>not</EM> account for that, which means that between 45 B.C. and
--    7 A.D. the leap years do not match the real years. (We correct the
--    Romans' counting error.)
--
--    Note that there was no year 0! The value 0 signifies 1 B.C., -1 is
--    2 B.C., and so on.</DL>
--  </PURPOSE>
--
--  <NOT_TASK_SAFE>
--
--  <NO_STORAGE>
--
--  <HISTORY>
--    03-OCT-2001   TW  Initial version.
--  </HISTORY>
-------------------------------------------------------------------------------

package body Util.Calendars.Julian is

   use Util.Times;
   use Util.Calendars.Western;

   function Is_Leap_Year
     (Base : in Calendar;
      Year : in Year_Number)
     return Boolean
   is
      pragma Warnings (Off, Base); --  silence -gnatwa

      Y : Year_Number := Year;
   begin
      --  We cannot take into account the Romans' counting error between
      --  45 B.C. and 8 A.D. because the julian day number conversion
      --  doesn't either.
      if Y < 0 then Y := -Y; end if;
      return Y mod 4 = 0;
   end Is_Leap_Year;

   ----------------------------------------------------------------------------
   --  Conversion routines.

   function Julian_Day
     (Base  : in Calendar;
      Year  : in Year_Number;
      Month : in Month_Number;
      Day   : in Day_Number)
     return Julian_Day_Number
   is
      pragma Warnings (Off, Base); --  silence -gnatwa
   begin
      if Year < -4712 then raise Date_Error; end if;
      --  Julian Day zero is Jan 1st, 4713 B.C (-4712), and since I haven't
      --  yet figured out how to extend the formula below to negative Julian
      --  Day numbers, this is our limit for the time being.
      declare
         A : Integer := (14 - Integer (Month)) / 12;
         Y : Integer := Integer (Year) + 4800 - A;
         M : Integer := Integer (Month) + 12 * A - 3;
      begin
         return
           Julian_Day_Number (Integer (Day) +
                              (153 * M + 2) / 5 +
                              Y * 365 + Y / 4 -
                              32083);
      end;
   end Julian_Day;

   procedure Date
     (Base       : in     Calendar;
      Julian_Day : in     Julian_Day_Number;
      Year       :    out Year_Number;
      Month      :    out Month_Number;
      Day        :    out Day_Number)
   is
      pragma Warnings (Off, Base); --  silence -gnatwa
   begin
      --  if Julian_Day < 0 then raise Date_Error; end if;
      declare
         C : Integer := Integer (Julian_Day) + 32082;
         D : Integer := (4 * C + 3) / 1461;
         E : Integer := C - (1461 * D) / 4;
         M : Integer := (5 * E + 2) / 153;
      begin
         Day   := Day_Number (E - (153 * M + 2) / 5 + 1);
         Month := Month_Number (M + 3 - 12 * (M / 10));
         Year  := Year_Number (D - 4800 + M / 10);
      end;
   end Date;

   ----------------------------------------------------------------------------
   --  Week days

   function Day_Of_Week
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Week_Day
   is
      Year  : Year_Number;
      Month : Month_Number;
      Day   : Day_Number;
   begin
      Split (Base, Date, Year, Month, Day);
      declare
         A : Integer := (14 - Integer (Month)) / 12;
         Y : Integer := Integer (Year) - A;
         M : Integer := Integer (Month) + 12 * A - 2;
         D : Integer := (5 + Integer (Day) +
                         Y + Y / 4 +
                         (31 * M) / 12) mod 7;
      begin
         --  Now D=0 means Sunday...
         D := (D + 6) mod 7;
         --  Now we've mapped it such that D=0 is Monday and D=6 is Sunday.
         return Week_Day'Val (D);
      end;
   end Day_Of_Week;

   ----------------------------------------------------------------------------
   --  Easter date

   function Easter
     (Base : in Calendar;
      Year : in Year_Number)
     return Util.Times.Time
   is
   begin
      if Year < 326 then raise Date_Error; end if;
      declare
         Y : Integer := Integer (Year);
         G : Integer := Y mod 19;
         I : Integer := (19 * G + 15) mod 30;
         J : Integer := (Y + Y / 4 + I) mod 7;
         L : Integer := I - J;
         M : Integer := 3 + (L + 40) / 44;
      begin
         return Time_Of (Base, Year, Month_Number (M),
                         Day_Number (L + 28 - 31 * (M / 4)));
      end;
   end Easter;

end Util.Calendars.Julian;
