-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2001, 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    This is a mixed proleptic Julian (for dates prior to and including
--    4-OCT-1582 A.D.) and Gregorian (for dates from 15-OCT-1582 on)
--    calendar, i.e. the calendar used in the western world, if we ignore
--    the chaotic differences in the times the Gregorian calendar was
--    actually adopted and the Romans' early counting error that caused every
--    <EM>third</EM> year from 45 B.C. to 8 B.C. be a leap year (which was
--    then corrected by not having any leap years at all from 7 B.C. until
--    8 A.D., which again was a leap year).
--
--    By the way, note that the day of the week did not change in Oct 1582,
--    Oct 4 was a Thursday, and Oct 15 was a Friday! (And so is, of course,
--    Oct 5, 1582 A.D. in the Julian calendar.)
--
--    Because not all countries and regions adopted the Gregorian calendar
--    in 1582 (notably many protestant areas switched much later, and some
--    orthodox regions didn't switch until the 20th century), you can define
--    the switch date. By default, it is set to 4-OCT-1582.</DL>
--  </PURPOSE>
--
--  <NOT_TASK_SAFE>
--
--  <NO_STORAGE>
--
--  <HISTORY>
--    03-OCT-2001   TW  Initial version
--    28-FEB-2002   TW  Added switch date. Note that this package does not
--                      support the pathological case of Sweden, which switched
--                      back and forth as they saw fit between the Julian and
--                      the Gregorian calendar between 1700 and 1753.
--  </HISTORY>
-------------------------------------------------------------------------------

with Util.Calendars.Western;
with Util.Calendars.Gregorian;
with Util.Calendars.Julian;
with Util.Times;

package body Util.Calendars.Universal is

   use Util.Times;

   J : Util.Calendars.Julian.Calendar;

   use Util.Calendars.Western;
   use Util.Calendars.Gregorian;
   use Util.Calendars.Julian;

   function Is_Leap_Year
     (Base : in Calendar;
      Year : in Year_Number)
     return Boolean
   is
   begin
      if Year <= Base.Switch_Year then
         return Is_Leap_Year (J, Year);
      else
         return Is_Leap_Year (Util.Calendars.Gregorian.Calendar (Base), Year);
      end if;
   end Is_Leap_Year;

   ----------------------------------------------------------------------------
   --  Conversion routines.

   function Julian_Day
     (Base  : in Calendar;
      Year  : in Year_Number;
      Month : in Util.Calendars.Western.Month_Number;
      Day   : in Util.Calendars.Western.Day_Number)
     return Util.Times.Julian_Day_Number
   is
   begin
      if Year <= Year_Number'Max (Base.Switch_Year, Base.Switch_Year_2) then
         declare
            Result : Julian_Day_Number := Julian_Day (J, Year, Month, Day);
         begin
            if Result > Base.Switch_Date then
               if Result <= Base.Switch_Date +
                            Base.Switch_Length +
                            Base.Switch_Length_2
               then
                  --  These dates do not exist
                  raise Date_Error;
               end if;
            else
               --  <= Base.Switch_Date: it *is* a Julian date.
               return Result;
            end if;
         end;
      end if;
      return Julian_Day (Util.Calendars.Gregorian.Calendar (Base),
                         Year, Month, Day);
   end Julian_Day;

   procedure Date
     (Base       : in     Calendar;
      Julian_Day : in     Util.Times.Julian_Day_Number;
      Year       :    out Year_Number;
      Month      :    out Util.Calendars.Western.Month_Number;
      Day        :    out Util.Calendars.Western.Day_Number)
   is
   begin
      if Julian_Day <= Base.Switch_Date then
         Date (J, Julian_Day, Year, Month, Day);
      else
         Date (Util.Calendars.Gregorian.Calendar (Base),
               Julian_Day, Year, Month, Day);
      end if;
   end Date;

   ----------------------------------------------------------------------------
   --  A system in which days are counted.

   function Day_In_Year
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Util.Calendars.Western.Day_Count
   is
   begin
      if Util.Times.Julian_Day (Date) <= Base.Switch_Date then
         return Day_In_Year (J, Date);
      else
         return Day_In_Year (Util.Calendars.Gregorian.Calendar (Base), Date);
      end if;
   end Day_In_Year;

   function Time_Of
     (Base    : in Calendar;
      Year    : in Year_Number;
      Day     : in Util.Calendars.Western.Day_Count;
      Seconds : in Util.Times.Day_Duration := 0.0)
     return Util.Times.Time
   is
   begin
      return
        Time_Of (Util.Calendars.Gregorian.Calendar (Base), Year, Day, Seconds);
   exception
      when Date_Error =>
         declare
            Nof_Days : Natural := 365;
         begin
            if Is_Leap_Year (J, Year) then Nof_Days := 366; end if;
            if Year = Base.Switch_Year and then
               Day > Nof_Days - Base.Switch_Length then
               raise;
            elsif Base.Switch_Year_2 /= 0 and then
                  Year = Base.Switch_Year_2 and then
                  Day > Nof_Days - Base.Switch_Length_2
            then
               raise;
            end if;
         end;
         return Time_Of (J, Year, Day, Seconds);
   end Time_Of;

   ----------------------------------------------------------------------------
   --  Easter date

   function Easter
     (Base : in Calendar;
      Year : in Year_Number)
     return Util.Times.Time
   is
   begin
      if Year <= Base.Switch_Year then
         declare
            Result : Util.Times.Time := Easter (J, Year);
         begin
            if Util.Times.Julian_Day (Result) <= Base.Switch_Date then
               return Result;
            end if;
         end;
      end if;
      return Easter (Util.Calendars.Gregorian.Calendar (Base), Year);
   end Easter;

   ----------------------------------------------------------------------------

   function Day_Of_Week
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Util.Calendars.Western.Week_Day
   is
   begin
      if Util.Times.Julian_Day (Date) <= Base.Switch_Date then
         return Day_Of_Week (J, Date);
      else
         return Day_Of_Week (Util.Calendars.Gregorian.Calendar (Base), Date);
      end if;
   end Day_Of_Week;

   procedure Set_Switch_Date
     (Base : in out Calendar;
      Day  : in     Util.Times.Time)
   is
   begin
      Set_Switch_Date (Base, Util.Times.Julian_Day (Day));
   end Set_Switch_Date;

   procedure Set_Switch_Date
     (Base : in out Calendar;
      Day  : in     Util.Times.Julian_Day_Number)
   is
      use Util.Times;

      Jul_Year     : Year_Number;
      Jul_Month    : Month_Number;
      Jul_Day      : Day_Number;
      Greg_Year    : Year_Number;
      Greg_Month   : Month_Number;
      Greg_Day     : Day_Number;
      X            : Natural;
   begin
      if Day < Base_Day then raise Date_Error; end if;
      --  Before that, there was no Gregorian calendar!
      Date (J, Day, Jul_Year, Jul_Month, Jul_Day);
      if Jul_Year > 2000 then raise Date_Error; end if;
      --  As far as I know, by 2001, no country is still using the Julian
      --  calendar. (For ecclesiastic date calculations, one may always use
      --  the Julian calendar, if needed.)
      Base.Switch_Date := Day;
      Base.Switch_Year := Jul_Year;
      Date (Util.Calendars.Gregorian.Calendar (Base),
            Day + 1, Greg_Year, Greg_Month, Greg_Day);
      Base.Switch_Length :=
        Natural (Julian_Day (J, Greg_Year, Greg_Month, Greg_Day) -
                 Day - 1);
      if Greg_Year > Base.Switch_Year then
         Base.Switch_Year_2 := Greg_Year;
         X := Natural (Julian_Day (J, Jul_Year, 12, 31) - Day);
         Base.Switch_Length_2 := Base.Switch_Length - X;
         Base.Switch_Length   := X;
      else
         Base.Switch_Year_2   := 0;
         Base.Switch_Length_2 := 0;
      end if;
   end Set_Switch_Date;

   function Switch_Date
     (Base : in Calendar)
     return Util.Times.Julian_Day_Number
   is
   begin
      return Base.Switch_Date;
   end Switch_Date;

end Util.Calendars.Universal;
