-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2001, 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <VERSION ID="1.1">
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    This is a mixed proleptic Julian (for dates prior to and including
--    4-OCT-1582 A.D.) and Gregorian (for dates from 15-OCT-1582 on)
--    calendar, i.e. the calendar used in the western world, if we ignore
--    the chaotic differences in the times the Gregorian calendar was
--    actually adopted and the Romans' early counting error that caused every
--    <EM>third</EM> year from 45 B.C. to 8 B.C. be a leap year (which was
--    then corrected by not having any leap years at all from 7 B.C. until
--    8 A.D., which again was a leap year).
--
--    By the way, note that the day of the week did not change in Oct 1582,
--    Oct 4 was a Thursday, and Oct 15 was a Friday! (And so is, of course,
--    Oct 5, 1582 A.D. in the Julian calendar.)
--
--    Because not all countries and regions adopted the Gregorian calendar
--    in 1582 (notably many protestant areas switched much later, and some
--    orthodox regions didn't switch until the 20th century), you can define
--    the switch date. By default, it is set to 4-OCT-1582.
--  </PURPOSE>
--
--  <NOT_TASK_SAFE>
--
--  <NO_STORAGE>
--
--  <HISTORY>
--    03-OCT-2001   TW  Initial version
--    28-FEB-2002   TW  Added switch date. Note that this package does not
--                      support the pathological case of Sweden, which switched
--                      back and forth as they saw fit between the Julian and
--                      the Gregorian calendar between 1700 and 1753.
--  </HISTORY>
-------------------------------------------------------------------------------

with Util.Calendars.Western;
with Util.Calendars.Gregorian;
with Util.Times;

package Util.Calendars.Universal is

   type Calendar is new Util.Calendars.Gregorian.Calendar with private;

   Date_Error : exception renames Util.Calendars.Date_Error;

   subtype Year_Number  is Util.Calendars.Gregorian.Year_Number;
   subtype Month_Number is Util.Calendars.Gregorian.Month_Number;
   subtype Day_Number   is Util.Calendars.Gregorian.Day_Number;

   function Is_Leap_Year
     (Base : in Calendar;
      Year : in Year_Number)
     return Boolean;

   ----------------------------------------------------------------------------
   --  Conversion routines.

   function Julian_Day
     (Base  : in Calendar;
      Year  : in Year_Number;
      Month : in Util.Calendars.Western.Month_Number;
      Day   : in Util.Calendars.Western.Day_Number)
     return Util.Times.Julian_Day_Number;

   procedure Date
     (Base       : in     Calendar;
      Julian_Day : in     Util.Times.Julian_Day_Number;
      Year       :    out Year_Number;
      Month      :    out Util.Calendars.Western.Month_Number;
      Day        :    out Util.Calendars.Western.Day_Number);

   ----------------------------------------------------------------------------
   --  A system in which days are counted.

   function Day_In_Year
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Util.Calendars.Western.Day_Count;

   function Time_Of
     (Base    : in Calendar;
      Year    : in Year_Number;
      Day     : in Util.Calendars.Western.Day_Count;
      Seconds : in Util.Times.Day_Duration := 0.0)
     return Util.Times.Time;

   ----------------------------------------------------------------------------
   --  Easter date

   function Easter
     (Base : in Calendar;
      Year : in Year_Number)
     return Util.Times.Time;

   ----------------------------------------------------------------------------

   function Day_Of_Week
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Util.Calendars.Western.Week_Day;

   ----------------------------------------------------------------------------

   procedure Set_Switch_Date
     (Base : in out Calendar;
      Day  : in     Util.Times.Time);
   --  @Day@ is assumed to be the last day the Julian calendar was in effect.
   --  Note that this package does <EM>not</EM> allow more than one switch date
   --  to be defined: the pathological case of Sweden, which switched back and
   --  forth between the Julian and the Gregorian calendar in a most irregular
   --  fashion between 1700 and 1753 is <EM>not</EM> handled.

   procedure Set_Switch_Date
     (Base : in out Calendar;
      Day  : in     Util.Times.Julian_Day_Number);
   --  @Day@ is assumed to be the last day the Julian calendar was in effect.

   function Switch_Date
     (Base : in Calendar)
     return Util.Times.Julian_Day_Number;

private

   Base_Day  : constant Util.Times.Julian_Day_Number := 2299160;
   --  04-OCT-1582 Julian.

   Base_Year : constant Year_Number                  := 1582;

   type Calendar is new Util.Calendars.Gregorian.Calendar with
      record
         Switch_Date     : Util.Times.Julian_Day_Number := Base_Day;
         --  Last day the gregorian calendar was in use.
         Switch_Year     : Year_Number  := Base_Year;
         Switch_Length   : Natural      := 10;
         Switch_Year_2   : Year_Number  := 0;
         Switch_Length_2 : Natural      := 0;
      end record;
   --  The "*_2" components come into effect if the switch cross new year's
   --  eve.

end Util.Calendars.Universal;
