-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2001, 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    Root package for western (christian) calendars.
--  </PURPOSE>
--
--  <NOT_TASK_SAFE>
--
--  <NO_STORAGE>
--
--  <HISTORY>
--    03-OCT-2001   TW  Initial version
--  </HISTORY>
-------------------------------------------------------------------------------

with Ada.Strings.Fixed;

package body Util.Calendars.Western is

   use Util.Times;

   function Year
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Year_Number
   is
      Y : Year_Number;
      M : Month_Number;
      D : Day_Number;
   begin
      Split (Calendar'Class (Base), Date, Y, M, D);
      return Y;
   end Year;

   function Month
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Month_Number
   is
      Y : Year_Number;
      M : Month_Number;
      D : Day_Number;
   begin
      Split (Calendar'Class (Base), Date, Y, M, D);
      return M;
   end Month;

   function Day
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Day_Number
   is
      Y : Year_Number;
      M : Month_Number;
      D : Day_Number;
   begin
      Split (Calendar'Class (Base), Date, Y, M, D);
      return D;
   end Day;

   procedure Split
     (Base    : in     Calendar;
      Date    : in     Util.Times.Time;
      Year    :    out Year_Number;
      Month   :    out Month_Number;
      Day     :    out Day_Number)
   is
   begin
      Util.Calendars.Western.Date
        (Calendar'Class (Base), Julian_Day (Date), Year, Month, Day);
   end Split;

   procedure Split
     (Base    : in     Calendar;
      Instant : in     Util.Times.Time;
      Year    :    out Year_Number;
      Month   :    out Month_Number;
      Day     :    out Day_Number;
      Seconds :    out Util.Times.Day_Duration)
   is
      JD : Julian_Day_Number;
   begin
      Split (Instant, JD, Seconds);
      Util.Calendars.Western.Date
        (Calendar'Class (Base), JD, Year, Month, Day);
   end Split;

   function Time_Of
     (Base    : in Calendar;
      Year    : in Year_Number;
      Month   : in Month_Number;
      Day     : in Day_Number;
      Seconds : in Util.Times.Day_Duration := 0.0)
     return Util.Times.Time
   is
   begin
      return
        Time_Of
          (Julian_Day (Calendar'Class (Base), Year, Month, Day), Seconds);
   end Time_Of;

   function Is_Leap_Year
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Boolean
   is
   begin
      return Is_Leap_Year (Calendar'Class (Base),
                           Year (Calendar'Class (Base), Date));
   end Is_Leap_Year;

   function Add_Months
     (Base   : in Calendar;
      Date   : in Util.Times.Time;
      Months : in Natural)
     return Util.Times.Time
   is
      Y : Year_Number;
      M : Month_Number;
      D : Day_Number;
   begin
      Split (Calendar'Class (Base), Date, Y, M, D);
      declare
         DM : Natural := Months mod 12;
      begin
         Y := Y + Year_Number (Months / 12);
         if Natural (M) + DM > 12 then
            Y := Y + 1;
            M := Month_Number (DM - (12 - Natural (M)));
         else
            M := M + DM;
         end if;
      end;
      return Time_Of (Calendar'Class (Base), Y, M, D, Util.Times.Secs (Date));
   end Add_Months;

   function Sub_Months
     (Base   : in Calendar;
      Date   : in Util.Times.Time;
      Months : in Natural)
     return Util.Times.Time
   is
      Y : Year_Number;
      M : Month_Number;
      D : Day_Number;
   begin
      Split (Calendar'Class (Base), Date, Y, M, D);
      declare
         DM : Natural := Months mod 12;
      begin
         Y := Y - Year_Number (Months / 12);
         --  Note the asymmetry in the condition in comparison to 'Add_Months'!
         if Natural (M) <= DM then
            Y := Y - 1;
            M := Month_Number (12 - DM + Natural (M));
         else
            M := Month_Number (Natural (M) - DM);
         end if;
      end;
      return Time_Of (Calendar'Class (Base), Y, M, D, Util.Times.Secs (Date));
   end Sub_Months;

   function Day_In_Year
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return Day_Count
   is
   begin
      return
        Day_Count (Julian_Day (Date) -
                   Julian_Day (Calendar'Class (Base),
                               Year (Calendar'Class (Base), Date),
                               1, 1) +
                   1);
   end Day_In_Year;

   function Time_Of
     (Base    : in Calendar;
      Year    : in Year_Number;
      Day     : in Day_Count;
      Seconds : in Day_Duration := 0.0)
     return Util.Times.Time
   is
   begin
      if Day = 366 and then not Is_Leap_Year (Calendar'Class (Base), Year) then
         raise Date_Error;
      end if;
      return
        Time_Of (Julian_Day (Calendar'Class (Base), Year, 1, 1) +
                 Julian_Day_Number (Day) - 1,
                 Seconds);
   end Time_Of;

   function Image
     (Base      : in Calendar;
      Date      : in Util.Times.Time;
      Format    : in Util.Calendar.IO.Date_Format;
      Separator : in String := "";
      Padding   : in Boolean := True;
      BC_AD     : in Boolean := False)
     return String
   is

      use type Util.Calendar.IO.Date_Format;

      function Trim
        (Source : in String;
         Side   : in Ada.Strings.Trim_End := Ada.Strings.Both)
        return String
        renames Ada.Strings.Fixed.Trim;
      --  Renaming to supply the 'Side' argument.

      function Canonical
        (Source  : in String;
         Padding : in Boolean)
        return String
      is
         function Pad
           (Source : in String;
            Count  : in Natural   := 2;
            Pad    : in Character := '0')
           return String
           renames Ada.Strings.Fixed.Tail;
         --  'Tail' is such a misleading name: we don't want to get at source's
         --  tail, we want to right-align and pad it with leading zeroes!

      begin --  Canonical
         if Padding then
            return Pad (Trim (Source));
         else
            return Trim (Source);
         end if;
      end Canonical;

      function BC_String
        (Format    : in Util.Calendar.IO.Date_Format;
         Separator : in String;
         BC        : in Boolean)
        return String
      is
      begin --  BC_String
         if not BC then
            return "";
         else
            if Format = YMD then
               return "BC" & Separator;
            else
               return " BC";
            end if;
         end if;
      end BC_String;

      Y : Year_Number;
      M : Month_Number;
      D : Day_Number;
      B : Boolean;

   begin --  Image
      Split (Calendar'Class (Base), Date, Y, M, D);
      B := BC_AD and then Y <= 0;
      if B then Y := -Y + 1; end if;
      declare
         YS : String := Trim (Year_Number'Image (Y));
         MS : String := Canonical (Month_Number'Image (M), Padding);
         DS : String := Canonical (Day_Number'Image (D), Padding);
         BC : String := BC_String (Format, Separator, B);
      begin
         case Format is
            when DMY =>
               return DS & Separator & MS & Separator & YS & BC;
            when MDY =>
               return MS & Separator & DS & Separator & YS & BC;
            when YMD =>
               return BC & YS & Separator & MS & Separator & DS;
         end case;
      end;
   end Image;

   function Image
     (Base : in Calendar;
      Date : in Util.Times.Time)
     return String
   is
   begin
      return Image (Calendar'Class (Base), Date, YMD, "-");
   end Image;

end Util.Calendars.Western;
