-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2001, 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <VERSION ID="1.1">
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    Stop watches; useful for simple timing tasks.
--  </PURPOSE>
--
--  <NOT_TASK_SAFE>
--
--  <NO_STORAGE>
--
--  <HISTORY>
--    28-FEB-2002   TW  Initial version.
--  </HISTORY>
-------------------------------------------------------------------------------

with Ada.Calendar;

package Util.Stop_Watches is

   pragma Elaborate_Body;

   type Stop_Watch is tagged private;
   --  @Tagged@ for two reasons:
   --
   --  <UL><LI>I want it to be a pass-by-reference type
   --          <BR><BR>
   --      <LI>I want to allow future extensions. One possible
   --          extension is to add a name to a stop watch. (See child package
   --          <CODE><A HREF="util-stop_watches-named.html">Named</A></CODE>.)
   --  </UL>
   --
   --  Any @Stop_Watch@ is initially reset to zero and isn't running.

   Stop_Watch_Error : exception;
   --  May be raised by @Start@ or @Elapsed@.

   procedure Reset
     (Watch : in out Stop_Watch);
   --  Stops a running stop watch and resets it to zero.

   procedure Start
     (Watch : in out Stop_Watch);
   --  Starts the stop watch. Raises @Stop_Watch_Error@ if the
   --  @Watch@ it is already running.

   procedure Stop
     (Watch : in out Stop_Watch);
   --  Stops a running stop watch. If the @Watch@ is not running,
   --  nothing happens.

   function Elapsed
     (Watch : in Stop_Watch)
     return Duration;
   --  Returns the accumulated elapsed time since the last @Reset@.
   --  Raises @Stop_Watch_Error@ if the @Watch@ is running.

   function Nof_Timings
     (Watch : in Stop_Watch)
     return Natural;
   --  Returns the number of timings, i.e., the number of calls to
   --  @Start@ that didn't raise an exception since the last
   --  call to @Reset@.

   function Average
     (Watch : in Stop_Watch)
     return Duration;
   --  Returns <CODE>Elapsed / Nof_Timings</CODE>; raises @Stop_Watch_Error@
   --  if @Nof_Timings@ is zero or if the @Watch@ is running.

private

   type Stop_Watch is tagged
      record
         Elapsed    : Duration := 0.0;
         Running    : Boolean  := False;
         Nof_Starts : Natural  := 0;
         Last_Start : Ada.Calendar.Time;
      end record;

end Util.Stop_Watches;
