-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    Provides a read-write lock type that fully supports lock promotion and
--    demotion, keeps track of the tasks currently holding a lock, and that
--    handles deserters. Conflicts arising when more than one task requests
--    a lock promotion from shared to exclusive mode can be handled through
--    a user-defined conflict resolution policy.
--
--    This type of lock is provided in a separate package because it
--    <EM>is</EM> rather complex, and we want to avoid dragging in all these
--    complexities when only the way simpler locks from package
--    <CODE><A HREF="util-tasking-locks.html">Util.Tasking.Locks</A></CODE>
--    are needed.
--  </PURPOSE>
--
--  <STORAGE>
--    Dynamic memory allocation in the standard storage pool(s).
--  </STORAGE>
--
--  <TASKING>
--    Fully task- and abortion-safe; deserters are handled.
--  </TASKING>
--
--  <HISTORY>
--    10-OCT-2002   TW  Initial version.
--  </HISTORY>
-------------------------------------------------------------------------------

with Util.Swap;

--  pragma Elaborate_All (Util.Swap);

package body Util.Tasking.Complex_Locks is

   use Ada.Task_Identification;

   ----------------------------------------------------------------------------

   procedure Finalize (Self : in out TRWL_Data)
   is
   begin
      if Self.Readers /= null  then Free (Self.Readers);  end if;
      if Self.Resolver /= null then Free (Self.Resolver); end if;
   end Finalize;

   function Find
     (List : in Task_List_Ptr;
      TID  : in Ada.Task_Identification.Task_Id)
     return Natural
   is
   begin
      if List /= null then
         for I in List'Range loop
            if List (I).TID = TID then return I; end if;
         end loop;
      end if;
      return 0;
   end Find;

   procedure Add_To_List
     (List      : in out Task_List_Ptr;
      TID       : in     Ada.Task_Identification.Task_Id;
      May_Exist : in     Boolean := False)
   is
      I : Natural := 0;
   begin
      if List = null then
         List := new Task_List'(1 => (TID        => TID,
                                      Promotion  => False,
                                      Lock_Times => 1));
      else
         if May_Exist then I := Find (List, TID); end if;
         if I > 0 then
            List (I).Lock_Times := List (I).Lock_Times + 1;
         else
            declare
               New_List : Task_List_Ptr := new Task_List (1 .. List'Last + 1);
            begin
               New_List (1 .. List'Last) := List.all;
               New_List (New_List'Last)  := (TID        => TID,
                                             Promotion  => False,
                                             Lock_Times => 1);
               Free (List);
               List := New_List;
            end;
         end if;
      end if;
   end Add_To_List;

   procedure Remove
     (List  : in out Task_List_Ptr;
      Index : in     Positive)
   is
   begin
      if List'Length = 1 then
         Free (List);
      else
         declare
            New_Readers : Task_List_Ptr;
         begin
            New_Readers := new Task_List (1 .. List'Length - 1);
            New_Readers (1 .. Index - 1) := List (1 .. Index - 1);
            New_Readers (Index .. New_Readers'Last) :=
              List (Index + 1 .. List'Last);
            Free (List);
            List := New_Readers;
         end;
      end if;
   end Remove;

   procedure Swap is new Util.Swap (Task_Descriptor);

   protected body Task_Read_Write_Lock is

      procedure Set_Conflict_Resolution
        (Resolver : in Conflict_Resolver'Class)
      is
      begin
         if Data.Resolver /= null then Free (Data.Resolver); end if;
         Data.Resolver :=
           new Conflict_Resolver'Class'(Resolver);
      end Set_Conflict_Resolution;

      procedure Set_Default_Conflict_Resolution
      is
      begin
         if Data.Resolver /= null then Free (Data.Resolver); end if;
      end Set_Default_Conflict_Resolution;

      --  All this pending change business is to make sure any queued up
      --  abandons, removes, and demotions get processed before any new
      --  gets and acquires, and that they trigger wait queue processing
      --  only when necessary.

      procedure Check_Pending_Change
        (Changed : in Boolean)
      is
      begin
         Data.Pending_Change := Data.Pending_Change or else Changed;
         if Abandon'Count = 0 and then
            Release'Count = 0 and then
            Demote'Count = 0 and then
            Data.Pending_Change
         then
            Data.Change := True; Data.Pending_Change := False;
         end if;
      end Check_Pending_Change;

      entry Get
        (Mode    : in     Lock_Mode;
         Granted :    out Boolean)
        when New_Promotion'Count = 0 and then
             Release'Count = 0 and then
             Abandon'Count = 0 and then
             Demote'Count = 0 and then
             not (Data.Change or else
                  Data.Pass_Through or else
                  Data.Do_Promotion)
      is
         Self : Natural := 0;
      begin
         if Data.Writer.TID = Get'Caller then
            --  This task itself is the writer: access granted.
            Granted := True;
            Data.Writer.Lock_Times := Data.Writer.Lock_Times + 1;
            return;
         end if;
         --  Check for deserters.
         if Data.Writer.TID /= Null_Task_Id and then
            Is_Terminated (Data.Writer.TID)
         then
            Data.Writer.TID        := Null_Task_Id;
            Data.Writer.Lock_Times := 0;
         end if;
         if Data.Readers /= null then
            for I in reverse Data.Readers'Range loop
               if Is_Terminated (Data.Readers (I).TID) then
                  if I < Self then Self := Self - 1; end if;
                  Remove (Data.Readers, I);
               elsif Data.Readers (I).TID = Get'Caller then
                  Self := I;
               end if;
            end loop;
         end if;
         Granted := Data.Writer.Lock_Times = 0;
         --  There's no writer.
         if Granted then
            if Self > 0 then
               --  I'm already a reader.
               if Mode = Shared then
                  Granted := True;
                  Data.Readers (Self).Lock_Times :=
                    Data.Readers (Self).Lock_Times + 1;
               else
                  --  A promotion request. This will only be granted if
                  --  this task is the only reader.
                  Granted := Data.Readers'Length = 1;
                  if Granted then
                     Data.Writer            := Data.Readers (Self);
                     Data.Writer.Lock_Times := Data.Writer.Lock_Times + 1;
                     Remove (Data.Readers, Self);
                  end if;
               end if;
            else
               Granted :=
                 Queued = 0 and then
                 (Mode = Shared or else Data.Readers = null);
               --  Nobody queued up, and it's a read request or there are no
               --  readers either. (We already know there's no writer.)
               if Granted then
                  if Mode = Shared then
                     Add_To_List (Data.Readers, Get'Caller);
                  else
                     Data.Writer.TID := Get'Caller;
                     Data.Writer.Lock_Times := 1;
                  end if;
               end if;
            end if;
         end if;
      end Get;

      entry Acquire
        (Mode : in Lock_Mode := Shared)
        when New_Promotion'Count = 0 and then
             Release'Count = 0 and then
             Abandon'Count = 0 and then
             Demote'Count = 0 and then
             not (Data.Change or else
                  Data.Pass_Through or else
                  Data.Do_Promotion)
      is
         Self : Natural := 0;
      begin
         if Data.Writer.TID = Acquire'Caller then
            --  Grant the request immediately.
            Data.Writer.Lock_Times := Data.Writer.Lock_Times + 1;
            return;
         end if;
         --  Check for deserters.
         if Data.Writer.TID /= Null_Task_Id and then
            Is_Terminated (Data.Writer.TID)
         then
            Data.Writer.TID        := Null_Task_Id;
            Data.Writer.Lock_Times := 0;
         end if;
         if Data.Readers /= null then
            for I in reverse Data.Readers'Range loop
               if Is_Terminated (Data.Readers (I).TID) then
                  if I < Self then Self := Self - 1; end if;
                  Remove (Data.Readers, I);
               elsif Data.Readers (I).TID = Acquire'Caller then
                  Self := I;
               end if;
            end loop;
         end if;
         if Self > 0 then
            if Mode = Exclusive then
               --  Lock promotion request.
               if Promote'Count > 0 then
                  --  There already is a promotion request
                  if Data.Resolver = null or else
                     Decide
                       (Data.Resolver,
                        Data.Readers (Data.Readers'First).TID, Acquire'Caller)
                  then
                     --  This new promotion request is refused.
                     raise Promotion_Error;
                  end if;
                  --  We need to kill an already queued promotion request.
                  Data.To_Kill := Data.Readers (Data.Readers'First).TID;
                  Data.Do_Promotion := True;
                  requeue New_Promotion; --  No "with abort" here!
                  --  Now, what'll happen is that first, the 'Promote' queue
                  --  opens, and the request queued there is killed. Next,
                  --  the 'New_Promotion' queu opens and moves the request over
                  --  to the (then empty) 'Promote' queue (which is closed: we
                  --  know that there are at least *two* readers here).
               else
                  if Data.Readers'Length = 1 then
                     --  Grant the promotion immediately. there cannot be a
                     --  writer, for then there wouldn't be any readers.
                     Data.Writer            := Data.Readers (Self);
                     Data.Writer.Lock_Times := Data.Writer.Lock_Times + 1;
                     Remove (Data.Readers, Self);
                     return;
                  else
                     --  A new promotion, and there isn't already one:
                     if Self > Data.Readers'First then
                        Swap (Data.Readers (Self),
                              Data.Readers (Data.Readers'First));
                     end if;
                     requeue Promote with abort;
                  end if;
               end if;
            else
               --  A reader asks for read-access again: grant it.
               Data.Readers (Self).Lock_Times :=
                 Data.Readers (Self).Lock_Times + 1;
               return;
            end if;
         end if;
         --  A regular request from a non-holder.
         Data.Change := Data.Writer.Lock_Times = 0;
         requeue Wait (Data.Active_Queue) with abort;
      end Acquire;

      entry New_Promotion
        when Data.To_Kill = Null_Task_Id and then Data.Do_Promotion
      is
         --  "Parking queue" to park the new promotion request until we've
         --  aborted the old one. There can be only one task on this queue!
         I : Natural := Find (Data.Readers, New_Promotion'Caller);
      begin
         --  Move the reader to the front, if it isn't there already.
         if I = 0 then raise Tasking_Error; end if; --  Can't happen!
         if I /= Data.Readers'First then
            Swap (Data.Readers (Data.Readers'First), Data.Readers (I));
         end if;
         Data.Do_Promotion := False;
         requeue Promote with abort;
      end New_Promotion;

      entry Promote
        when Data.To_Kill /= Null_Task_Id or else Data.Do_Promotion
      is
         --  There can be only one single task on this queue!
      begin
         if Promote'Caller = Data.To_Kill then
            Data.To_Kill := Null_Task_Id;
            raise Promotion_Error;
         else
            --  We finally got our promotion!
            if Data.Readers (Data.Readers'First).TID /= Promote'Caller then
               raise Program_Error; --  Can't happen!
            end if;
            Data.Writer            := Data.Readers (Data.Readers'First);
            Data.Writer.Lock_Times := Data.Writer.Lock_Times + 1;
            Free (Data.Readers);
            Data.Do_Promotion := False;
         end if;
      end Promote;

      entry Wait (for Queue in Boolean) (Mode : in Lock_Mode)
        when Queue = Data.Active_Queue and then
             (not Data.Do_Promotion) and then
             (Data.Change or else Data.Pass_Through)
      is
      begin
         --  We only have non-holders on this queue!
         if Data.Pass_Through then
            --  Just move the waiting tasks over to the other queue
            if Wait (Queue)'Count = 0 then
               --  Last one: stop passing through, wait for the next change,
               --  and switch the active queue.
               Data.Pass_Through := False;
               Data.Change       := False;
               Data.Active_Queue := not Queue;
            end if;
            requeue Wait (not Queue) with abort;
         else
            if Mode = Shared then
               if Data.Writer.Lock_Times = 0 then
                  --  Grant read-only access
                  Add_To_List (Data.Readers, Wait'Caller, True);
                  if Wait (Queue)'Count = 0 then Data.Change := False; end if;
               else
                  Data.Change := False;
                  Data.Pass_Through := Wait (Queue)'Count > 0;
                  if not Data.Pass_Through then
                     --  No need to switch queues.
                     requeue Wait (Queue) with abort;
                  else
                     requeue Wait (not Queue) with abort;
                  end if;
               end if;
            else
               --  A writer. It cannot be a lock promotion! (Because we have
               --  only non-holders on this queue.)
               Data.Change := False;
               if Data.Readers = null and then Data.Writer.Lock_Times = 0 then
                  --  No readers, and no writer either: grant write access.
                  Data.Writer.TID        := Wait'Caller;
                  Data.Writer.Lock_Times := Data.Writer.Lock_Times + 1;
                  --  A writer just got the lock: no need to pass through,
                  --  for it must have been the very first request on the
                  --  queue.
                  Data.Pass_Through := False;
               else
                  Data.Pass_Through := Wait (Queue)'Count > 0;
                  --  Oops, there are readers, or another writer.
                  if not Data.Pass_Through then
                     requeue Wait (Queue) with abort;
                  else
                     requeue Wait (not Queue) with abort;
                  end if;
               end if;
            end if; --  Reader or writer?
         end if; --  Pass_Through?
      end Wait;

      entry Abandon
        when not (Data.Change or else Data.Pass_Through or else
                  Data.To_Kill /= Null_Task_Id)
      is
      begin
         if Data.Writer.TID = Abandon'Caller then
            Data.Writer.Lock_Times := 0;
            Data.Writer.TID        := Null_Task_Id;
         else
            declare
               I : Natural := Find (Data.Readers, Abandon'Caller);
            begin
               if I = 0 then raise Tasking_Error; end if;
               Data.Do_Promotion :=
                 Data.Readers'Length = 2 and then Promote'Count > 0;
               --  Note: if the length is 2 now, it'll be 1 after the remove,
               --  and that remaining reader will be the one queued on the
               --  'Promote' entry. Checking here avoids an additional null-
               --  check, though, for after 'Remove', 'Data.Readers' may be
               --  null.
               Remove (Data.Readers, I);
            end;
         end if;
         Check_Pending_Change (Wait (Data.Active_Queue)'Count > 0 and then
                               Promote'Count = 0);
      end Abandon;

      entry Release
         when not (Data.Change or else Data.Pass_Through or else
                   Data.To_Kill /= Null_Task_Id)
      is
      begin
         if Data.Writer.TID = Release'Caller then
            Data.Writer.Lock_Times := Data.Writer.Lock_Times - 1;
            if Data.Writer.Lock_Times = 0 then
               Data.Writer.TID := Null_Task_Id;
            end if;
         else
            declare
               I : Natural := Find (Data.Readers, Release'Caller);
            begin
               if I = 0 then raise Tasking_Error; end if;
               Data.Readers (I).Lock_Times := Data.Readers (I).Lock_Times - 1;
               if Data.Readers (I).Lock_Times = 0 then
                  Data.Do_Promotion :=
                    Data.Readers'Length = 2 and then Promote'Count > 0;
                  --  Note: if the length is 2 now, it'll be 1 after the
                  --  remove, and that remaining reader will be the one
                  --  queued on the 'Promote' entry.
                  Remove (Data.Readers, I);
               end if;
            end;
         end if;
         Check_Pending_Change (Wait (Data.Active_Queue)'Count > 0 and then
                               Promote'Count = 0);
      end Release;

      entry Demote
        when not (Data.Change or else Data.Pass_Through)
      is
      begin
         if Data.Writer.TID = Demote'Caller then
            Add_To_List (Data.Readers, Data.Writer.TID);
            --  It's the last entry in now.
            Data.Readers (Data.Readers'Last).Lock_Times :=
              Data.Writer.Lock_Times;
            --  Give up the write lock:
            Data.Writer.Lock_Times := 0;
            Data.Writer.TID        := Null_Task_Id;
            Check_Pending_Change (Wait (Data.Active_Queue)'Count > 0);
            --  There cannot be a queued promotion request: we were the writer
            --  until just now!
         else
            Check_Pending_Change (False);
            --  No change here, but an earlier call might have changed
            --  something
            if Find (Data.Readers, Demote'Caller) = 0 then
               raise Tasking_Error;
            end if;
         end if;
      end Demote;

      function  Queued
        return Natural
      is
      begin
         return Acquire'Count + Wait (False)'Count + Wait (True)'Count +
                Promote'Count;
      end Queued;

      function  Writer return Ada.Task_Identification.Task_Id
      is
      begin
         return Data.Writer.TID;
      end Writer;

      entry Holding (Mode : out Lock_Holder_Mode)
        when not Data.Change
      is
      begin
         Mode := Holds (Holding'Caller);
      end Holding;

      function Holds
        (TID : in Ada.Task_Identification.Task_Id)
        return Lock_Holder_Mode
      is
      begin
         if TID = Data.Writer.TID then return Exclusive; end if;
         declare
            I : Natural := Find (Data.Readers, TID);
         begin
            if I > 0 then
               if Data.Readers (I).Promotion then
                  return Shared_Exclusive;
               end if;
               return Shared;
            end if;
         end;
         return Not_Holder;
      end Holds;

   end Task_Read_Write_Lock;

   ----------------------------------------------------------------------------

   procedure Initialize
     (Self : in out Task_Read_Write_Lock_Protector)
   is
   begin
      Self.Lock.Acquire (Self.Mode);
   end Initialize;

   procedure Finalize
     (Self : in out Task_Read_Write_Lock_Protector)
   is
   begin
      if Self.Lock.Holds (Current_Task) /= Not_Holder then
         Self.Lock.Release;
      end if;
   exception
      when Tasking_Error =>
         --  Ooops, something went really haywire.
         null;
   end Finalize;

end Util.Tasking.Complex_Locks;
