-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    Various useful lock types. Note that in general it's better to
--    encapsulate your shared data properly in protected objects than using
--    explicit lock POs, which always are in some sense and abstraction
--    inversion. Code using explicit locks may be prone to suffer from
--    priority inversions. However, there are situations in which you just
--    cannot use PO encapsulation, e.g. when one of the protected operations
--    should do potentially blocking calls such as any I/O. In such cases, you
--    need to use explicit locks, and this package provides the most commonly
--    needed ones.
--
--    Also see package
--    <CODE><A HREF="util-tasking-protectors.html">Util.Tasking.Protectors</A>
--    </CODE>
--    for some controlled types that help using these locks in an exception-
--    and ATC-safe way.
--  </PURPOSE>
--
--  <NO_STORAGE>
--
--  <TASKING>
--    Fully task- and abortion-safe. Only the @Task_Lock@ handles deserters;
--    other lock types don't, and it's the application's business to ensure
--    that such locks are properly released even in the presence of exceptions
--    and ATC.
--  </TASKING>
--
--  <HISTORY>
--    10-OCT-2002   TW  Initial version.
--  </HISTORY>
-------------------------------------------------------------------------------

package body Util.Tasking.Locks is

   protected body Lock is

      procedure Get (Granted : out Boolean)
      is
      begin
         Granted := not Locked;
         --  Note: contrary to the other lock types below, it is not necessary
         --  here to check for Acquire'Count = 0: since entry queues are
         --  serviced as part of the protected operation 'Release', any queued
         --  requests on 'Acquire' will be serviced anyway before a call to
         --  'Get'.
         if Granted then Locked := True; end if;
      end Get;

      entry     Acquire
        when not Locked
      is
      begin
         Locked := True;
      end Acquire;

      procedure Release
      is
      begin
         Locked := False;
      end Release;

      function  Queued return Natural
      is
      begin
         return Acquire'Count;
      end Queued;

   end Lock;

   ----------------------------------------------------------------------------

   protected body Read_Write_Lock is
      --  In general, reader-writer locks are prone to starvation. Without
      --  special provisions, a stream of readers could starve a writer
      --  (i.e., when readers arrive at a sufficiently high rate such that
      --  the count of readers never drops back to zero). We therefore requeue
      --  readers and writers onto the same queue, and hence readers cannot
      --  overtake writers of equal priority. The only way some task might be
      --  starved is then that it gets passed by all the time by requests from
      --  tasks with a higher priority, which is consistent with the Ada 95
      --  rules.

      procedure Get
        (Read_Write : in     Boolean;
         Granted    :    out Boolean)
      is
      begin
         Granted :=
           Wait (Active_Queue)'Count = 0 and then
           not Writer and then (Readers = 0 or else not Read_Write);
         --  Here, we *do* need to check for waiting requests, for otherwise
         --  read-only calls to 'Get' might bypass queued requests. Note that
         --  if we have queued requests, the front one is a read-write request
         --  anyway.
         if Granted then
            if Read_Write then
               Writer := True;
            else
               Readers := Readers + 1;
            end if;
         end if;
      end Get;

      entry Acquire (Read_Write : Boolean := False)
        when not Pass_Through
      is
         --  We block incoming acquire calls while we're passing through
         --  entries from one queue to the other.
      begin
         if not Writer then
            Change := True;
            --  Always open the active queue, even if there are already waiting
            --  tasks. The reason is that this request might have a higher
            --  priority than any of the tasks already waiting. Even if this
            --  request might block, it still might jump to the head of the
            --  queue; and if it's a read-only request, it might even be
            --  serviced if the current queue head is a read-write request
            --  that is blocked because there currently are readers.
         end if;
         --  If we have a writer, we'll block anyway.
         requeue Wait (Active_Queue) with abort;
      end Acquire;

      procedure Release  is
      begin
         if Readers > 0 then
            --  This call to 'Release' is assumed to have come from a reader.
            Readers := Readers - 1;
            if Readers = 0 then
               Change := Queued > 0;
            end if;
         elsif Writer then
            --  Call from a writer.
            Writer := False;
            Change := Queued > 0;
         end if;
      end Release;

      entry Wait (for Queue in Boolean) (Read_Write : in Boolean)
        when Queue = Active_Queue and then
             ((not Writer and then Change) or else Pass_Through)
      is
      begin
         if Pass_Through then
            --  Just move the waiting tasks over to the other queue
            if Wait (Queue)'Count = 0 then
               --  Last one: stop passing through, wait for the next change,
               --  and switch the active queue.
               Pass_Through := False; Change := False;
               Active_Queue := not Queue;
            end if;
            requeue Wait (not Queue) with abort;
         else
            if Read_Write then
               --  In any case, wait for the next change: either this writer
               --  gets the lock and shuts out all other requests, or it
               --  doesn't get the lock and therefore this request itself plus
               --  all others behind it remain blocked.
               Change := False;
               if Readers = 0 then
                  --  Writer gets the lock: no passing through; the active
                  --  queue is still the current one.
                  Writer := True;
               else
                  --  There are readers: the writer and any task behind it
                  --  remain blocked.
                  Pass_Through := Wait (Queue)'Count > 0;
                  if not Pass_Through then
                     --  No further tasks: active queue remains this one.
                     requeue Wait (Queue) with abort;
                  else
                     requeue Wait (not Queue) with abort;
                  end if;
               end if;
            else
               --  We only get here if Writer = False. A new reader enters.
               Readers := Readers + 1;
               if Wait (Queue)'Count = 0 then
                  --  We had only readers in the queue.
                  Change := False;
               end if;
            end if;
         end if;
      end Wait;

      function Queued return Natural
      is
      begin
         return Acquire'Count + Wait (False)'Count + Wait (True)'Count;
      end Queued;

   end Read_Write_Lock;

   ----------------------------------------------------------------------------

   use Ada.Task_Identification;

   protected body Task_Lock is
      --  Note: this lock type *does* handle deserters by checking each time
      --  a request is made to get the lock while it is held that the holding
      --  task is not terminated.

      entry Get
        (Granted : out Boolean)
        when True
      is
      begin
         if Owner /= Null_Task_Id and then Is_Terminated (Owner) then
            --  Oops, a deserter!
            Owner := Null_Task_Id;
            Count := 0;
         end if;
         Granted :=
           Wait'Count = 0 and then (Count = 0 or else Owner = Get'Caller);
         if Granted then
            Owner := Get'Caller;
            Count := Count + 1;
         end if;
      end Get;

      entry Acquire
        when True
      is
      begin
         if Owner = Acquire'Caller then
            --  The lock already is ours!
            Count := Count + 1;
         else
            if Owner /= Null_Task_Id and then Is_Terminated (Owner) then
               --  Oops, a deserter!
               Owner := Null_Task_Id;
               Count := 0;
            end if;
            requeue Wait with abort;
         end if;
      end Acquire;

      entry Abandon
        when True
      is
      begin
         --  'Abandon' is an entry to allow use of the 'Caller attribute.
         if Count > 0 then
            if Owner /= Abandon'Caller then raise Tasking_Error; end if;
            Count := 0;
            Owner := Null_Task_Id;
         end if;
      end Abandon;

      entry Release
        when True
      is
      begin
         --  'Release' is an entry to allow use of the 'Caller attribute.
         if Count > 0 then
            if Owner /= Release'Caller then raise Tasking_Error; end if;
            Count := Count - 1;
            if Count = 0 then Owner := Null_Task_Id; end if;
         end if;
      end Release;

      entry Wait
        when Count = 0
      is
      begin
         --  One task can only be queued once on @Wait@.
         Owner := Wait'Caller;
         Count := 1;
      end Wait;

      function Queued return Natural is
      begin
         return Acquire'Count + Wait'Count;
         --  Actually, I think Acquire'Count is always zero.
      end Queued;

      function Holder
        return Ada.Task_Identification.Task_Id
      is
      begin
         return Owner;
      end Holder;

   end Task_Lock;

end Util.Tasking.Locks;


