-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    Root package of the tasking utilities subsystem. Provides two generic
--    packages for making arbitrary (constrained, non-limited) types task-safe.
--  </PURPOSE>
--
--  <NO_STORAGE>
--
--  <TASKING>
--    Fully task- and abortion-safe.
--  </TASKING>
--
--  <HISTORY>
--    10-OCT-2002   TW  Initial version.
--  </HISTORY>
-------------------------------------------------------------------------------

package body Util.Tasking is

   --  generic
   --     type Item is private;
   --     Default : in Item;
   package body Protect is

      protected body Protected_Item is

         procedure Set (To : in Item)
         is
         begin
            Value := To;
         end Set;

         function  Get return Item
         is
         begin
            return Value;
         end Get;

         procedure Set_And_Get
           (New_Value : in     Item;
            Old_Value :    out Item)
         is
         begin
            Old_Value := Value; Value := New_Value;
         end Set_And_Get;

      end Protected_Item;

   end Protect;

   ----------------------------------------------------------------------------

   --  generic
   --     type Item is private;
   --     Default : in Item;
   --     function "=" (Left, Right : in Item) return Boolean is <>;
   package body Protect_With_Wait is

      protected body Protected_Item is

         procedure Set (To : in Item)
         is
         begin
            Change := Value /= To;
            Value  := To;
         end Set;

         function  Get return Item
         is
         begin
            return Value;
         end Get;

         procedure Set_And_Get
           (New_Value : in     Item;
            Old_Value :    out Item)
         is
         begin
            Old_Value := Value; Value := New_Value;
            Change := Old_Value /= New_Value;
         end Set_And_Get;

         entry Wait (Val : in Item)
           when True
         is
         begin
            if Value /= Val then
               requeue Wait_Until (Active_Queue) with abort;
            end if;
         end Wait;

         entry Wait_Until (for Queue in Boolean) (Val : in Item)
           when Queue = Active_Queue and then Change
         is
         begin
            if Wait_Until (Queue)'Count = 0 then
               --  Last one: switch queues, and wait for the next change
               Change := False; Active_Queue := not Queue;
            end if;
            if Value /= Val then
               requeue Wait_Until (not Queue) with abort;
            end if;
         end Wait_Until;

      end Protected_Item;

   end Protect_With_Wait;

end Util.Tasking;
