-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2001, 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    A generic swap routine.
--  </PURPOSE>
--
--  <NOT_TASK_SAFE>
--
--  <NO_STORAGE>
--
--  <HISTORY>
--    28-FEB-2002   TW  Initial version.
--  </HISTORY>
-------------------------------------------------------------------------------

with Ada.Text_IO;
with Ada.Strings.Fixed;

with Util.Calendar.IO;

package body Util.Times.IO is

   ----------------------------------------------------------------------------
   --  Utility functions

   function Pad
     (Source : in String;
      Count  : in Natural   := 2;
      Pad    : in Character := '0')
     return String
     renames Ada.Strings.Fixed.Tail;
   --  'Tail' is such a misleading name: we don't want to get at source's
   --  tail, we want to right-align and pad it with leading zeroes!

   function Trim
     (Source : in String;
      Side   : in Ada.Strings.Trim_End := Ada.Strings.Both)
     return String
     renames Ada.Strings.Fixed.Trim;
   --  Renaming to supply the 'Side' argument.

   ----------------------------------------------------------------------------
   --  Images for type 'Day_Duration'

   function Image
     (Instant      : in Day_Duration;
      With_Seconds : in Boolean := True;
      AM_PM        : in Boolean := False)
     return String
     renames Util.Calendar.IO.Image;

   function Image
     (Instant   : in Day_Duration;
      Precision : in Natural;
      AM_PM     : in Boolean := False)
     return String
     renames Util.Calendar.IO.Image;

   ----------------------------------------------------------------------------
   --  Images for type 'Time_Difference'

   type Hours_Type is mod 2 ** 64;

   function Image
     (Instant   : in Time_Difference;
      With_Days : in Boolean := True)
     return String
   is
   begin
      if Instant.D < 0 or else Instant.S < 0.0 then
         return '-' & Image (-Instant, With_Days);
      end if;
      if With_Days then
         declare
            S : String := Image (Instant.S, True) & 's';
         begin
            S (S'First + 2) := 'h';
            S (S'First + 5) := 'm';
            return Trim (Julian_Day_Number'Image (Instant.D)) & 'd' & S;
         end;
      else
         declare
            H  : Hours;
            M  : Minutes;
            S  : Seconds;
            F  : Fraction;
            HH : Hours_Type;
         begin
            Split (Instant.S, H, M, S, F);
            HH := Hours_Type (H) + Hours_Type (Instant.D) * 24;
            return Trim (Hours_Type'Image (HH)) & 'h' &
                   Pad (Trim (Minutes'Image (M))) & 'm' &
                   Pad (Trim (Seconds'Image (S))) & 's';
         end;
      end if;
   end Image;

   package HIO is new Ada.Text_IO.Modular_IO (Hours_Type);

   function Image
     (Instant   : in Time_Difference;
      Precision : in Natural;
      With_Days : in Boolean := True)
     return String
   is
   begin
      if Instant.D < 0 or else Instant.S < 0.0 then
         return '-' & Image (-Instant, With_Days);
      end if;
      declare
         S : String := Image (Instant.S, Precision) & 's';
      begin
         S (S'First + 2) := 'h';
         S (S'First + 5) := 'm';
         if With_Days then
            if S (S'First .. S'First + 1) = "24" then
               S (S'First .. S'First + 1) := "00";
               return Trim (Julian_Day_Number'Image (Instant.D + 1)) & 'd' & S;
            else
               return Trim (Julian_Day_Number'Image (Instant.D)) & 'd' & S;
            end if;
         else
            declare
               HH   : Hours_Type;
               Last : Positive;
            begin
               HIO.Get (S (S'First .. S'First + 1), HH, Last);
               HH := HH + Hours_Type (Instant.D) * 24;
               return Trim (Hours_Type'Image (HH)) &
                      S (S'First + 2 .. S'Last);
            end;
         end if;
      end;
   end Image;

end Util.Times.IO;
