-------------------------------------------------------------------------------
--
--  <STRONG>Copyright &copy; 2001, 2002 by Thomas Wolf.</STRONG>
--  <BLOCKQUOTE>
--    This piece of software is free software; you can redistribute it and/or
--    modify it under the terms of the  GNU General Public License as published
--    by the Free Software  Foundation; either version 2, or (at your option)
--    any later version. This software is distributed in the hope that it will
--    be useful, but <EM>without any warranty</EM>; without even the implied
--    warranty of <EM>merchantability or fitness for a particular purpose.</EM>
--    See the GNU General Public License for  more details. You should have
--    received a copy of the GNU General Public License with this distribution,
--    see file "<A HREF="GPL.txt">GPL.txt</A>". If not, write to the Free
--    Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
--    USA.
--  </BLOCKQUOTE>
--  <BLOCKQUOTE>
--    As a special exception from the GPL, if other files instantiate generics
--    from this unit, or you link this unit with other files to produce an
--    executable, this unit does not by itself cause the resulting executable
--    to be covered by the GPL. This exception does not however invalidate any
--    other reasons why the executable file might be covered by the GPL.
--  </BLOCKQUOTE>
--
--  <AUTHOR>
--    Thomas Wolf  (TW) <E_MAIL>
--  </AUTHOR>
--
--  <PURPOSE>
--    A generic swap routine.
--  </PURPOSE>
--
--  <NOT_TASK_SAFE>
--
--  <NO_STORAGE>
--
--  <HISTORY>
--    28-FEB-2002   TW  Initial version.
--    28-OCT-2002   TW  Added 'UTC' and 'Is_DST'.
--  </HISTORY>
-------------------------------------------------------------------------------

with Ada.Calendar;

with Util.Calendar.UTC;

with Util.Calendars.Gregorian;
with Util.Calendars.Western;

package body Util.Times is

   subtype Year_N is Util.Calendars.Year_Number;

   ----------------------------------------------------------------------------
   --  Conversion routines.

   function To_Time
     (Instant : in Ada.Calendar.Time)
     return Time
   is
      A_Y : Ada.Calendar.Year_Number;
      A_M : Ada.Calendar.Month_Number;
      A_D : Ada.Calendar.Day_Number;
      S   : Ada.Calendar.Day_Duration;
      G   : Util.Calendars.Gregorian.Calendar;
   begin
      Ada.Calendar.Split (Instant, A_Y, A_M, A_D, S);
      return Util.Calendars.Gregorian.Time_Of
               (G, Year_N (A_Y), A_M, A_D, S);
   end To_Time;

   function To_Time
     (Instant : in Time)
     return Ada.Calendar.Time
   is
      Y : Year_N;
      M : Util.Calendars.Western.Month_Number;
      D : Util.Calendars.Western.Day_Number;
      S : Day_Duration;
      G : Util.Calendars.Gregorian.Calendar;
   begin
      Util.Calendars.Gregorian.Split (G, Instant, Y, M, D, S);
      if Y < Year_N (Ada.Calendar.Year_Number'First) or else
         Y > Year_N (Ada.Calendar.Year_Number'Last)
      then
         raise Time_Error;
      end if;
      begin
         return
           Ada.Calendar.Time_Of (Ada.Calendar.Year_Number (Y),
                                 Ada.Calendar.Month_Number (M),
                                 Ada.Calendar.Day_Number (D),
                                 Ada.Calendar.Day_Duration (S));
      exception
         when others =>
            raise Time_Error;
      end;
   end To_Time;

   ----------------------------------------------------------------------------
   --  Accessors

   function Now
     return Time
   is
   begin
      return To_Time (Ada.Calendar.Clock);
   end Now;

   function UTC
     return Time
   is
   begin
      return To_Time (Util.Calendar.UTC.UTC);
   end UTC;

   function Is_DST
     return Boolean
     renames Util.Calendar.UTC.Is_DST;

   function Secs
     (Instant : in Time)
     return Day_Duration
   is
   begin
      return Instant.S;
   end Secs;

   function Julian_Day
     (Instant : in Time)
     return Julian_Day_Number
   is
   begin
      return Instant.JD;
   end Julian_Day;

   procedure Split
     (Instant    : in     Time;
      Julian_Day :    out Julian_Day_Number;
      Seconds    :    out Day_Duration)
   is
   begin
      Julian_Day := Instant.JD;
      Seconds    := Instant.S;
   end Split;

   procedure Split
     (Secs : in     Day_Duration;
      Hrs  :    out Hours;
      Min  :    out Minutes;
      Sec  :    out Seconds)
   is
      S : Natural := Natural (Secs); --  Rounds
   begin
      Hrs := Hours (S / 3600);
      S   := S mod 3600;
      Min := Minutes (S / 60);
      Sec := Seconds (S mod 60);
   end Split;

   procedure Split
     (Secs : in     Day_Duration;
      Hrs  :    out Hours;
      Min  :    out Minutes;
      Sec  :    out Seconds;
      Frac :    out Fraction)
   is
      F : Day_Duration;
   begin
      Util.Calendar.Split (Secs, Hrs, Min, Sec, F);
      Frac := F;
   end Split;

   function To_Duration
     (Hrs  : in Hours;
      Min  : in Minutes;
      Sec  : in Seconds;
      Frac : in Fraction := 0.0)
     return Day_Duration
   is
   begin
      return Day_Duration (Hrs) * 3600.0 +
             Day_Duration (Min) * 60.0 +
             Day_Duration (Sec) + Day_Duration (Frac);
   end To_Duration;

   ----------------------------------------------------------------------------
   --  Constructors

   function Time_Of
     (Julian_Day : in Julian_Day_Number;
      Seconds    : in Day_Duration := 0.0)
     return Time
   is
   begin
      if Seconds = Day_Duration'Last then
         return Time'(JD => Julian_Day + 1, S => 0.0);
      else
         return Time'(JD => Julian_Day, S => Seconds);
      end if;
   end Time_Of;

   ----------------------------------------------------------------------------
   --  Operations on Time values.

   function "+"
     (Date : in Time;
      Days : in Julian_Day_Number)
     return Time
   is
   begin
      return Time'(JD => Date.JD + Days, S => Date.S);
   end "+";

   function "+"
     (Days : in Julian_Day_Number;
      Date : in Time)
     return Time
   is
   begin
      return Time'(JD => Date.JD + Days, S => Date.S);
   end "+";

   function "-"
     (Date : in Time;
      Days : in Julian_Day_Number)
     return Time
   is
   begin
      return Time'(JD => Date.JD - Days, S => Date.S);
   end "-";

   function Days_Between
     (L, R : in Time)
     return Julian_Day_Number
   is
   begin
      return L.JD - R.JD;
   end Days_Between;

   function "+"
     (Instant : in Time;
      Offset  : in Day_Duration)
     return Time
   is
      Result : Time := Instant;
   begin
      if Day_Duration'Last - Offset < Instant.S then
         --  We'll get an overflow.
         Result.JD := Result.JD + 1;
         Result.S  := Offset - (Day_Duration'Last - Result.S);
      else
         Result.S := Result.S + Offset;
      end if;
      return Instant;
   end "+";

   function "+"
     (Offset  : in Day_Duration;
      Instant : in Time)
     return Time
   is
   begin
      return Instant + Offset;
   end "+";

   function "-"
     (Instant : in Time;
      Offset  : in Day_Duration)
     return Time
   is
      Result : Time := Instant;
   begin
      if Offset > Result.S then
         Result.JD := Result.JD - 1;
         Result.S  := Day_Duration'Last - (Offset - Result.S);
      else
         Result.S := Result.S - Offset;
      end if;
      return Result;
   end "-";

   function "<"
     (L, R : in Time)
     return Boolean
   is
   begin
      return L.JD < R.JD or else
             (L.JD = R.JD and then L.S < R.S);
   end "<";

   function "<="
     (L, R : in Time)
     return Boolean
   is
   begin
      return not (L > R);
   end "<=";

   function ">"
     (L, R : in Time)
     return Boolean
   is
   begin
      return L.JD > R.JD or else
             (L.JD = R.JD and then L.S > R.S);
   end ">";

   function ">="
     (L, R : in Time)
     return Boolean
   is
   begin
      return not (L < R);
   end ">=";

   ----------------------------------------------------------------------------
   --  Time Differences

   --  Unary operators

   function "+"
     (Value : in Time_Difference)
     return Time_Difference
   is
   begin
      return Value;
   end "+";

   function "-"
     (Value : in Time_Difference)
     return Time_Difference
   is
   begin
      return (-Value.D, -Value.S);
   end "-";

   --  Binary operators

   function "+"
     (Left, Right : in Time_Difference)
     return Time_Difference
   is
      Result : Time_Difference;
   begin
      Result.D := Left.D + Right.D;
      if Left.S <= 0.0 and then Right.S <= 0.0 and then
         -Left.S >= Day_Duration'Last + Right.S
      then
         Result.D := Result.D - 1;
         Result.S := (Day_Duration'Last + Left.S) + Right.S;
      elsif Left.S >= 0.0 and then Right.S >= 0.0 and then
            Left.S >= Day_Duration'Last - Right.S
      then
         Result.D := Result.D + 1;
         Result.S := (Left.S - Day_Duration'Last) + Right.S;
      else
         Result.S := Left.S + Right.S;
      end if;
      return Result;
   end "+";

   function "-"
     (Left, Right : in Time_Difference)
     return Time_Difference
   is
   begin
      return Left + (-Right);
   end "-";

   function "*"
     (Left  : in Time_Difference;
      Right : in Integer)
     return Time_Difference
   is
      Result : Time_Difference;
   begin
      case Right is
         when -2 => return -(Left + Left);
         when -1 => return Time_Difference'(-Left.D, -Left.S);
         when  0 => return Time_Difference'(0, 0.0);
         when  1 => return Left;
         when  2 => return Left + Left;
         when others =>
            null;
      end case;
      Result.D := Left.D * Right;
      if Left.S = 0.0 then
         Result.S := 0.0;
      else
         if (Left.S > 0.0 and then Left.S > Day_Duration'Last / abs Right)
            or else
            (Left.S < 0.0 and then -Left.S > Day_Duration'Last / abs Right)
         then
            --  We'll get an overflow. How can we compute (without overflowing)
            --  by how many days we'll overflow? I don't know, and hence this
            --  recursive argument reduction, which'll allow us to do the whole
            --  multiplication with a minimum of additions.
            declare
               L : Integer := Right / 2;
               R : Integer := Right - L;
            begin
               return Left * L + Left * R;
            end;
         else
            --  No overflow.
            Result.S := Left.S * Right;
            if Result.S = Day_Duration'Last then
               Result.S := 0.0;
               Result.D := Result.D + 1;
            elsif Result.S = -Day_Duration'Last then
               Result.S := 0.0;
               Result.D := Result.D - 1;
            end if;
         end if;
      end if;
      return Result;
   end "*";

   function "*"
     (Left  : in Integer;
      Right : in Time_Difference)
     return Time_Difference
   is
   begin
      return Right * Left;
   end "*";

   function "*"
     (Left  : in Time_Difference;
      Right : in Float)
     return Time_Difference
   is
   begin
      return From_Float (To_Float (Left) * Right);
   end "*";

   function "*"
     (Left  : in Float;
      Right : in Time_Difference)
     return Time_Difference
   is
   begin
      return Right * Left;
   end "*";

   function "/"
     (Left, Right : in Time_Difference)
     return Integer
   is
      F : Float;
   begin
      F := Left / Right;
      return Integer (Float'Truncation (F));
   end "/";

   function "/"
     (Left, Right : in Time_Difference)
     return Float
   is
   begin
      return To_Float (Left) / To_Float (Right);
   end "/";

   function "/"
     (Left  : in Time_Difference;
      Right : in Integer)
     return Time_Difference
   is
   begin
      return Left / Float (Right);
   end "/";

   function "/"
     (Left  : in Time_Difference;
      Right : in Float)
     return Time_Difference
   is
   begin
      return From_Float (To_Float (Left) / Right);
   end "/";

   --  Relational operators

   function "<"
     (Left, Right : in Time_Difference)
     return Boolean
   is
   begin
      return Left.D < Right.D or else
             (Left.D = Right.D and then Left.S < Right.S);
   end "<";

   function "<="
     (Left, Right : in Time_Difference)
     return Boolean
   is
   begin
      return not (Left > Right);
   end "<=";

   function ">"
     (Left, Right : in Time_Difference)
     return Boolean
   is
   begin
      return Left.D > Right.D or else
             (Left.D = Right.D and then Left.S > Right.S);
   end ">";

   function ">="
     (Left, Right : in Time_Difference)
     return Boolean
   is
   begin
      return not (Left < Right);
   end ">=";

   --  Conversions

   function To_Float
     (T : in Time_Difference)
     return Float
   is
   begin
      return Float (T.D) * 86_400.0 + Float (T.S);
   end To_Float;

   function From_Float
     (F : in Float)
     return Time_Difference
   is
      Result : Time_Difference;
   begin
      Result.D := Julian_Day_Number (Float'Truncation (F / 86_400.0));
      Result.S := Duration (F - Float (Result.D) * 86_400.0);
      if abs Result.S = Day_Duration'Last then
         if Result.S < 0.0 then
            Result.D := Result.D - 1;
         else
            Result.D := Result.D + 1;
         end if;
         Result.S := 0.0;
      end if;
      return Result;
   end From_Float;

   --  Constructor

   function To_Duration
     (Days    : in Julian_Day_Number;
      Seconds : in Day_Duration := 0.0)
     return Time_Difference
   is
   begin
      if Seconds = Day_Duration'Last then
         return Time_Difference'(Days + 1, 0.0);
      else
         return Time_Difference'(Days, Seconds);
      end if;
   end To_Duration;

   --  Accessors

   procedure Split
     (Secs : in     Time_Difference;
      Neg  :    out Boolean;
      Days :    out Julian_Day_Number;
      Hrs  :    out Hours;
      Min  :    out Minutes;
      Sec  :    out Seconds)
   is
      S : Time_Difference := Secs;
   begin
      if S.D < 0 or else S.S < 0.0 then
         Neg := True;
         S.D := abs S.D;
         S.S := abs S.S;
      else
         Neg := False;
      end if;
      Days := S.D;
      Split (S.S, Hrs, Min, Sec);
   end Split;

   procedure Split
     (Secs : in     Time_Difference;
      Neg  :    out Boolean;
      Days :    out Julian_Day_Number;
      Hrs  :    out Hours;
      Min  :    out Minutes;
      Sec  :    out Seconds;
      Frac :    out Fraction)
   is
      S : Time_Difference := Secs;
   begin
      if S.D < 0 or else S.S < 0.0 then
         Neg := True;
         S.D := abs S.D;
         S.S := abs S.S;
      else
         Neg := False;
      end if;
      Days := S.D;
      Split (S.S, Hrs, Min, Sec, Frac);
   end Split;

   --  Operations on 'Time' values

   function "-"
     (L, R : in Time)
     return Time_Difference
   is
   begin
      if L < R then
         return -(R - L);
      elsif L = R then
         return Zero;
      end if;
      --  L > R, so the result is positive.
      declare
         Result : Time_Difference;
      begin
         Result.D := L.JD - R.JD;
         if Result.D = 0 then
            --  Same day.
            Result.S := L.S - R.S;
         else
            if R.S > L.S then
               --  Less than a day
               Result.D := Result.D - 1;
               Result.S := L.S + (86_400.0 - R.S);
            else
               Result.S := L.S - R.S;
            end if;
         end if;
         return Result;
      end;
   end "-";

   function "+"
     (Instant : in Time;
      Offset  : in Time_Difference)
     return Time
   is
      Result : Time;
   begin
      if Offset.D < 0 or else Offset.S < 0.0 then
         return Instant - (-Offset);
      end if;
      Result.JD := Instant.JD + Offset.D;
      if Offset.S >= 86_400.0 - Instant.S then
         Result.JD := Result.JD + 1;
         Result.S  := Offset.S - (86_400.0 - Instant.S);
      else
         Result.S := Instant.S + Offset.S;
      end if;
      return Result;
   end "+";

   function "+"
     (Offset  : in Time_Difference;
      Instant : in Time)
     return Time
   is
   begin
      return Instant + Offset;
   end "+";

   function "-"
     (Instant : in Time;
      Offset  : in Time_Difference)
     return Time
   is
      Result : Time;
   begin
      if Offset.D < 0 or else Offset.S < 0.0 then
         return Instant + (-Offset);
      end if;
      Result.JD := Instant.JD - Offset.D;
      if Instant.S < Offset.S then
         Result.JD := Result.JD - 1;
         Result.S  := 86_400.0 - (Offset.S - Instant.S);
      else
         Result.S := Instant.S - Offset.S;
      end if;
      return Result;
   end "-";

end Util.Times;


