/******************************************************************************

  Copyright (c) 2001, 2002 by Thomas Wolf.

  This piece of software is free software; you can redistribute it and/or
  modify it under the terms of the  GNU General Public License as published
  by the Free Software  Foundation; either version 2, or (at your option)
  any later version. This unit is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
  Public License for  more details. You should have received a copy of the
  GNU General Public License with this distribution, see file GPL.txt. If not,
  write to the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
  MA 02111-1307, USA.

  As a special exception from the GPL, if you link this unit with other files
  to produce an executable, this unit does not by itself cause the resulting
  executable to be covered by the GPL. This exception does not however
  invalidate any other reasons why the executable file might be covered by
  the GPL.

  Author:
    Thomas Wolf  (TW)   twolf@acm.org

  Purpose:
    Small C wrapper functions for the standard ISO-C library.

  Storage semantics:
    No storage allocation.

  Revision History

    29-APR-2002  TW  Initial version.
    28-OCT-2002  TW  Added 'get_utc'.
 *****************************************************************************/

/*
  Small C routine to portably get the value of '\n'. (A '\n' in a string
  is replaced at *compile-time*, so there's no portable way to obtain its
  value from Ada 95. And note also that the ISO C standard leaves the actual
  character value of '\n' as well as its external representation in a file
  implementation-defined. Thus we cannot rely on '\n' being the linefeed
  character (Character'Val (10)), nor can we assume that it would appear
  as such in a file (it might just as well end up as a CR-LF or a single
  CR in the file). Hence this minor work-around.)
*/
int util_ada_get_nl (void)
{
  return '\n';
}

/*
  Routine for getting the current time in UTC. We cannot use time_t in Ada
  programs because the ISO C standard defines time_t as an "arithmetic type
  capable of representing times": it might be long on some systems, long long
  on others, or even double. I don't want to have to hack around to configure
  the Util subsystem depending on the host it runs on, so there's this little
  wrapper that allows us to get a broken down UTC time without having to use
  time_t anywhere in the Ada part.
    Furthermore, note that "struct tm" is defined minimally by the ISO C
  standard as a structure containing "at least the following members, in any
  order", so we'd even have a hard time defining an Ada record for it. See
  ISO 9899:1990, 7.12.1, or ISO 9899:1999, 7.23.1 (3,4).
    Finally, note that neither the C90 nor the C99 standard have any way to
  get at a sub-second fraction!
    This function returns -1 if 'utc_time' is NULL or UTC time is not
  available. If it returns zero, the structure pointed to by 'utc_time' has
  been filled in with the broken down current UTC time.
 */

#include <time.h>

typedef struct {
          int year;     /* >= 1900 */
          int month;    /* 1 .. 12 */
          int day;      /* 1 .. 31 */
          int hours;    /* 0 .. 23 */
          int minutes;  /* 0 .. 59 */
          int seconds;  /* 0 .. 61; 60 .. 61 for leap seconds?! */
        } C_Time_Type;

int util_ada_get_utc (C_Time_Type *utc_time)
{
  time_t    now;
  struct tm *utc;

  if (!utc_time) return -1; /* Shouldn't happen. */
  time (&now);
  utc = gmtime (&now);
  if (!utc) return -1; /* UTC could not be determined. */
  utc_time->year    = utc->tm_year + 1900;
  /* gmtime returns years since 1900. */
  utc_time->month   = utc->tm_mon + 1;     /* gm_time returns 0 .. 11! */
  utc_time->day     = utc->tm_mday;
  utc_time->hours   = utc->tm_hour;
  utc_time->minutes = utc->tm_min;
  utc_time->seconds = utc->tm_sec;
  return 0;
}

/*
  Another small helper function. Returns zero if not DST, > 0 if DST, and < 0
  if unknown.
 */
int util_ada_is_dst (void)
{
  time_t    now;
  struct tm *local_time;

  time (&now);
  local_time = localtime (&now);
  if (!local_time) return -1;
  return local_time->tm_isdst;
}



